import atexit
import json
import arcgis

_DASHBOARD_VERSION = 27

_created_dashboards = []


class Dashboard(object):
    """
    Creates a Dashboard Object.

    :return Dashboard object
    """
    def __init__(self):
        self.elements = []

        self._theme = 'light'

        self._header = None
        self._side_panel = None
        self._layout = {}
        self._widgets = []

        # if item is not None:
        #     return _from_dashboard(item)
        
    def save(self, title, description='', summary='', tags=None, gis=None, overwrite=False):
        """
        Publishes a Dashboard Object.

        =========================   ===========================================
        **Argument**                **Description**
        -------------------------   -------------------------------------------
        title                       Required string. Title or Caption for the
                                    Dashboard.
        -------------------------   -------------------------------------------
        description                 Optional string. Description for the Dashboard.
        -------------------------   -------------------------------------------
        summary                     Optional string. Summary of the Dashboard.
        -------------------------   -------------------------------------------
        tags                        Optional string. Comma separated tags.
        -------------------------   -------------------------------------------
        gis                         Optional GIS to publish dashboard.
                                    By default uses active gis.
        -------------------------   -------------------------------------------
        overwrite                   Optional Boolean.
                                    Overwrite existing dashboard.
        =========================   ===========================================
        """
        if not title:
            raise Exception("Please specify a title.")

        self.title = title
        self.description = description

        self.summary = summary
        self.tags = tags if tags else ''
        return self._publish(gis=gis, overwrite=overwrite)

    @property
    def theme(self):
        return self._theme

    @theme.setter
    def theme(self, value):
        self._theme = value
        if value not in ["light", "dark"]:
            self._theme = "light"

    @property
    def header(self):
        """
        :return: Header Object
        """
        return self._header

    @header.setter
    def header(self, value):
        """
        Set the header object
        """
        self._header = value

    @property
    def side_panel(self):
        """
        :return Side Panel Object
        """
        return self._side_panel

    @side_panel.setter
    def side_panel(self, value):
        """
        Set the Side Panel object
        """
        self._side_panel = value

    def _convert_to_json(self):
        json_data = {
            "version": _DASHBOARD_VERSION,
            "widgets":
                [
                    element._convert_to_json()
                    for element in self.elements
                ],
            "settings":
                {
                    "maxPaginationRecords": 50000,
                    "allowElementResizing": False
                },
            "mapOverrides":
                {
                    "trackedFeatureRadius": 60
                },
            "theme": self.theme,
            "themeOverrides": {},
            "numberPrefixOverrides": [],
            "layout": self._layout,
            "authoringApp": "ArcGIS API for Python",
            "authoringAppVersion": arcgis.__version__,
            "typeKeywords": "ArcGIS Dashboard(via Python)"
        }

        if self.header:
            json_data['headerPanel'] = self.header._convert_to_json()

        if self.side_panel:
            json_data['leftPanel'] = self.side_panel._convert_to_json()

        # print(json_data)
        return json_data

    @property
    def layout(self):
        """
        :return: Layout of the dashboard
        """
        return self._layout

    @layout.setter
    def layout(self, value):
        """
        Set the layout of the dashboard, using add_row and add_column functions.
        """
        self.elements = value['widgets']
        del value['widgets']
        self._layout = {"rootElement" : value}

    # @property
    # def widgets(self):
    #     """
    #     :return: widgets of the dashboard
    #     """
    #     return self._widgets
    
    def _repr_html_(self):
        url = self._dash_publish()
        return f"""<iframe src={url} width=900 height=300>"""

    def _dash_publish(self):
        gis = arcgis.env.active_gis
        import random
        import string

        letters = string.ascii_lowercase
        title = ''.join(random.choice(letters) for i in range(10))
        summary = ''.join(random.choice(letters) for i in range(10))

        db = Dashboard()

        db.title = title
        db.summary = summary
        db.description = ''
        db.summary = summary
        db.tags = ''

        db._layout = self._layout
        db.elements = self.elements
        db = db._publish(gis)
        _created_dashboards.append((gis, db))
        url = f'{gis.url}/apps/opsdashboard/index.html#/{db.itemid}'

        return url

    def _publish(self, gis=None, overwrite=False):
        if gis is None:
            gis = arcgis.env.active_gis

        items = gis.content.search(f'title:{self.title}', 'Dashboard')
        for item in items:
            if item.title.lower() == self.title.lower() and overwrite is False:
                raise Exception("A dashboard with same name already exists, to continue set `overwrite` = True")
            elif item.title.lower() == self.title.lower():
                item.delete(force=True)

        return gis.content.add({
            'type': 'Dashboard',
            'description': self.description,
            'title': self.title,
            'overwrite': str(overwrite).lower(),
            'text': json.dumps(self._convert_to_json())}
        )

    def _from_dashboard(self, dashboard_item):
        widget_map = {
            "indicatorWidget": arcgis.apps.dashboard.Indicator,
            "gaugeWidget": arcgis.apps.dashboard.Gauge,
            "pieChartWidget": arcgis.apps.dashboard.PieChart,
            "serialChartWidget": arcgis.apps.dashboard.SerialChart,
            "detailsWidget": arcgis.apps.dashboard.Details,
            "richTextWidget": arcgis.apps.dashboard.RichText,
            "listWidget": arcgis.apps.dashboard.List,
            "embeddedContentWidget": arcgis.apps.dashboard.EmbeddedContent,
            "legendWidget": arcgis.apps.dashboard.MapLegend
        }
        item_json = dashboard_item.get_data()
        for widget_json in item_json["widgets"]:
            type = widget_json["type"]
            self._widgets.append(widget_map[type]._from_json(widget_json))

        return dashboard_item

    @staticmethod
    def _publish_random(widget):
        from arcgis.apps.dashboard import add_row
        gis = arcgis.env.active_gis
        import random
        import string

        letters = string.ascii_lowercase
        title = ''.join(random.choice(letters) for i in range(10))
        summary = ''.join(random.choice(letters) for i in range(10))
        db = Dashboard()

        if widget.type == 'headerPanel':
            db.header = widget
        elif widget.type == 'leftPanel':
            db.side_panel = widget
        elif widget.type == 'legendWidget':
            map_widget_width = widget._map_widget.width
            widget._map_widget.width = 0.5
            db.layout = add_row([widget._map_widget, widget])
            widget._map_widget.width = map_widget_width
        else:
            if hasattr(widget, 'item') and getattr(widget.item, 'type', None) == 'mapWidget':
                db.layout = add_row([widget.item, widget])
            else:
                db.layout = add_row([widget])

        db = db.save(title, summary, gis=gis, overwrite=True)
        _created_dashboards.append((gis, db))
        url = f'{gis.url}/apps/opsdashboard/index.html#/{db.itemid}'

        return url


@atexit.register
def _delete_objects():
    for dashboard in _created_dashboards:
        gis = dashboard[0]
        db = dashboard[1]
        db.delete(force=True)


import re
import os
import sys
import copy
from copy import deepcopy
import types
import random
import warnings
import traceback
import pandas as pd
from functools import partial

HAS_FASTAI = True
try:
    from enum import Enum
    import torch
    import arcgis   
    from fastprogress.fastprogress import progress_bar
    from fastai.text import TextList, TextClasDataBunch, pad_collate, TextDataBunch,\
         SortishSampler, SortSampler, ItemList, ItemBase, Text
    from fastai.text import *
    from fastai.layers import CrossEntropyFlat
    from fastai.basic_train import Learner
    from fastai.torch_core import to_data
    from fastai.callback import Callback
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False

HAS_NUMPY = True
try:
    import numpy as np
    warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
except:
    HAS_NUMPY = False


max_len=100
DatasetType = Enum('DatasetType', 'Train Valid Test Single Fix')

class SequenceToSequenceDataBunch(TextDataBunch):

    "Create a `TextDataBunch` suitable for training."
    @classmethod
    def create(cls, train_ds, valid_ds, test_ds=None, path='.', bs=32, pad_idx=1,
               dl_tfms=None, pad_first=False, no_check=False, 
               backwards=False, val_bs=None, collate_fn=None, **dl_kwargs):
        "Function that transform the `datasets` in a `DataBunch`. Passes `**dl_kwargs` on to `DataLoader()`"
        device = dl_kwargs.pop('device',None)
        from torch.utils.data import DataLoader
        datasets = cls._init_ds(train_ds, valid_ds)
        val_bs = bs
        collate_fn = partial(seq2seq_collate, pad_idx=pad_idx, pad_first=pad_first, backwards=backwards)
        train_sampler = SortishSampler(datasets[0].x, key=lambda t: len(datasets[0][t][0].data), bs=bs//2)
        train_dl = DataLoader(datasets[0], batch_size=bs, sampler=train_sampler, drop_last=True, **dl_kwargs)
        dataloaders = [train_dl]
        for i,ds in enumerate(datasets[1:]):
            lengths = [len(t) for t in ds.x.items]
            sampler = SortSampler(ds.x, key=lengths.__getitem__)
            dataloaders.append(DataLoader(ds, batch_size=val_bs, sampler=sampler,  **dl_kwargs))
        return cls(*dataloaders, dl_tfms=dl_tfms, path=path, collate_fn=collate_fn, no_check=no_check, device=device)

class SequenceToSequenceTextList(TextList):
    _bunch = SequenceToSequenceDataBunch
    _label_cls = TextList

    def reconstruct(self, t:torch.Tensor):
        
        if isinstance(t,(list)): # handling the 'shift_tfm' dataloader transform
            t=t[0]
        elif t.ndim==2: 
            t = t.argmax(-1)
        idx_min = (t != self.pad_idx).nonzero().min()
        idx_max = (t != self.pad_idx).nonzero().max()
        return Text(t[idx_min:idx_max+1], self.vocab.textify(t[idx_min:idx_max+1]))

class SequenceToSequenceLearner(Learner):
    def __init__(self, data,  model, metrics=None, loss_func=None, **learn_kwargs):
        # self.model_name = model_name
        pretrained_model_name = model.pretrained_model_name
        super().__init__(data, model, metrics=metrics, loss_func=loss_func, **learn_kwargs)
        self.tokenizer = self.model._tokenizer #AutoTokenizer.from_pretrained(pretrained_model_name)
        if self.tokenizer.__class__.__name__ == 'MarianTokenizer':
            self.tokenizer.clean_up_tokenization = clean_up_tokenization


    def show_results(self, ds_type=DatasetType.Valid, rows:int=5, num_beams=4, max_len=50):
        if not HAS_NUMPY:
            raise Exception("This function requires numpy.")
        from IPython.display import display, HTML
        "Show `rows` result of predictions on `ds_type` dataset."
        ds = self.dl(ds_type).dataset
        x,y = self.data.one_batch(ds_type, detach=False, denorm=False)
        preds = self.pred_batch(batch=(x,y))
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(rows)]
        ys = [ds.x.reconstruct(grab_idx(y, i)) for i in range(rows)]
        zs = self.model._transformer.generate(x[0], num_beams=num_beams, max_length=max_len)
        items,names = [],['text', 'target', 'pred']
        for i, (x,y,z) in enumerate(zip(xs,ys,zs)):          
            txt_x = self.tokenizer.decode(x.data, skip_special_tokens=True)
            txt_y = self.tokenizer.decode(y.data, skip_special_tokens=True)
            txt_z = self.tokenizer.decode(z, skip_special_tokens=True)
            items.append([txt_x, txt_y, txt_z])
        items = np.array(items)
        df = pd.DataFrame({n:items[:,i] for i,n in enumerate(names)}, columns=names)
        with pd.option_context('display.max_colwidth', -1):
            display(HTML(df.to_html(index=False)))
    
    def predict_batch(self, batch_text, num_beams, max_length, min_length):
        tok = self.model._tokenizer
        encoded_input_batch = tok.batch_encode_plus(batch_text, padding=True, return_tensors='pt')['input_ids']
        encoded_output_batch =  self.model._transformer.generate(encoded_input_batch.to(self.model._transformer.device.type),\
                                                                 num_beams=num_beams, max_length=max_length, min_length=min_length)
        decoded_output_batch = tok.batch_decode(encoded_output_batch, skip_special_tokens=True)
        return decoded_output_batch, encoded_output_batch


    def predict(self, text_list, batch_size, show_progress, **kwargs):# num_beams=4, max_len=50):
        tok = self.model._tokenizer
        num_beams = kwargs.get('num_beams', 1)
        max_length = kwargs.get('max_length', 20)
        min_length = kwargs.get('min_length', 10)

        
        if isinstance(text_list,(list)):
            decoded_output = []
            for i in progress_bar(range(0, len(text_list), batch_size), display=show_progress):
                batch_text = text_list[i:i+batch_size]
                decoded_output_batch, encoded_output_batch = self.predict_batch(batch_text, \
                                                                num_beams=num_beams, max_length=max_length, min_length=min_length)
                decoded_output.extend(decoded_output_batch)
        else:
            encoded_input = text_list
            # encoded_input.unsqueeze_(0)
            encoded_output =  self.model._transformer.generate(encoded_input.to(self.model._transformer.device.type),\
                                                             num_beams=num_beams, max_length=max_length, min_length=min_length)
            decoded_output = tok.batch_decode(encoded_output, skip_special_tokens=True)
        return decoded_output

class NGram():
    if not HAS_NUMPY:
            raise Exception("This function requires numpy.")
    def __init__(self, ngram, max_n=5000): self.ngram,self.max_n = ngram,max_n
    def __eq__(self, other):
        if len(self.ngram) != len(other.ngram): return False
        return np.all(np.array(self.ngram) == np.array(other.ngram))
    def __hash__(self): return int(sum([o * self.max_n**i for i,o in enumerate(self.ngram)]))

def get_grams(x, n, max_n=5000):
    return x if n==1 else [NGram(x[i:i+n], max_n=max_n) for i in range(len(x)-n+1)]

def get_correct_ngrams(pred, targ, n, max_n=5000):
    pred_grams,targ_grams = get_grams(pred, n, max_n=max_n),get_grams(targ, n, max_n=max_n)
    pred_cnt,targ_cnt = Counter(pred_grams),Counter(targ_grams)
    return sum([min(c, targ_cnt[g]) for g,c in pred_cnt.items()]),len(pred_grams)

class CorpusBLEU(Callback):
    def __init__(self, vocab_sz):
        self.vocab_sz = vocab_sz
        self.name = 'bleu'
    
    def on_epoch_begin(self, **kwargs):
        self.pred_len,self.targ_len,self.corrects,self.counts = 0,0,[0]*4,[0]*4
    
    def on_batch_end(self, last_output, last_target, **kwargs):
        loss, last_output = last_output
        last_output = last_output.argmax(dim=-1)
        for pred,targ in zip(last_output.cpu().numpy(),last_target.cpu().numpy()):
            self.pred_len += len(pred)
            self.targ_len += len(targ)
            for i in range(4):
                c,t = get_correct_ngrams(pred, targ, i+1, max_n=self.vocab_sz)
                self.corrects[i] += c
                self.counts[i]   += t
    
    def on_epoch_end(self, last_metrics, **kwargs):
        precs = [c/t for c,t in zip(self.corrects,self.counts)]
        len_penalty = exp(1 - self.targ_len/self.pred_len) if self.pred_len < self.targ_len else 1
        bleu = len_penalty * ((precs[0]*precs[1]*precs[2]*precs[3]) ** 0.25)
        return add_metrics(last_metrics, bleu)
  
def seq2seq_loss(out, targ, pad_idx=1):
    loss, _ = out
    return loss

def seq2seq_acc(out, targ, pad_idx=1):
    loss, out = out
    bs,targ_len = targ.size()
    _,out_len,vs = out.size()
    if targ_len>out_len: out  = F.pad(out,  (0,0,0,targ_len-out_len,0,0), value=pad_idx)
    if out_len>targ_len: targ = F.pad(targ, (0,out_len-targ_len,0,0), value=pad_idx)
    out = out.argmax(2)
    return (out==targ).float().mean()

def seq2seq_collate(samples, pad_idx=1, pad_first=True, backwards=False):
    "Function that collect samples and adds padding. Flips token order if needed"
    samples = to_data(samples)
    max_len_x,max_len_y = max([len(s[0]) for s in samples]),max([len(s[1]) for s in samples])
    res_x = torch.zeros(len(samples), max_len_x).long() + pad_idx
    res_y = torch.zeros(len(samples), max_len_y).long() + pad_idx
    if backwards: pad_first = not pad_first
    for i,s in enumerate(samples):
        if pad_first: 
            res_x[i,-len(s[0]):],res_y[i,-len(s[1]):] = torch.LongTensor(s[0]),torch.LongTensor(s[1])
        else:         
            res_x[i, :len(s[0])],res_y[i, :len(s[1])] = torch.LongTensor(s[0]),torch.LongTensor(s[1])
    if backwards: res_x,res_y = res_x.flip(1),res_y.flip(1)
    return res_x, res_y

# def shift_tfm(b):
#     #transform to modify input for teacher forcing
#     x,y = b
#     y = torch.nn.functional.pad(y, (1, 0), value=0)
#     return [x,y[:,:-1]], y[:,1:]

def teacher_forcing_tfm(b):
    x,y = b
    return [x,y], y

def clean_up_tokenization(out_string: str) -> str:
        """
        Additional handling of '▁' in clean_up_tokenization for marian tokenizer.
        """
        out_string = (
            out_string.replace(" .", ".")
            .replace(" ?", "?")
            .replace(" !", "!")
            .replace(" ,", ",")
            .replace(" ' ", "'")
            .replace(" n't", "n't")
            .replace(" 'm", "'m")
            .replace(" 's", "'s")
            .replace(" 've", "'ve")
            .replace(" 're", "'re")
        )
        out_string = re.sub(r'▁(?=\S)', ' ',out_string)
        return out_string
    




import json
import logging
import random
import traceback
from functools import partial
from pathlib import Path
import glob
import warnings
import os

from arcgis.learn._data import _raise_fastai_import_error
from ._arcgis_model import ArcGISModel
from ._codetemplate import image_classifier_prf

HAS_OPENCV = True
HAS_FASTAI = True
HAS_ARCPY = True

# Try to import the necessary modules
# Exception will turn the HAS_FASTAI flag to false
# so that relevant exception can be raised
try:
    import numpy as np
    from fastai.vision.learner import create_body
    from torch import optim
    from torchvision import models
    from arcgis.learn.models._linknet_utils import (
        LinkNetMultiTaskModel,
        accuracy,
        miou,
        road_orient_loss,
    )
    from arcgis.learn.models._hourglass_utils import StackHourglassMultiTaskModel
    from ._road_mtl_learner import MultiTaskRoadLearner
    from arcgis.learn.models._arcgis_model import _resnet_family, _EmptyData
    from arcgis.learn._utils.common import (
        get_multispectral_data_params_from_emd ,_get_emd_path
    )
    from arcgis.learn._utils.segmentation_loss_functions import dice
    from arcgis.learn.models._arcgis_model import _device_check
    from .._data_utils._pixel_classifier_data import ClassifiedTilesData
    from .._data_utils._road_orient_data import RoadOrientation
except Exception as e:
    import_exception = "\n".join(
        traceback.format_exception(type(e), e, e.__traceback__)
    )
    HAS_FASTAI = False

def safe_json(data):
    """
    Check if data is JSON Serializable.
    It can be used while saving emd_parameters, as it dumps
    the data using JSON which requires JSON serializanble data only.
    """
    if data is None:
        return True
    elif isinstance(data, (bool, int, float)):
        return True
    elif isinstance(data, (tuple, list)):
        return all(safe_json(x) for x in data)
    elif isinstance(data, dict):
        return all(isinstance(k, str) and safe_json(v) for k, v in data.items())
    return False

class MultiTaskRoadExtractor(ArcGISModel):
    """
    Creates a Multi-Task Learning model for binary segmentation. Supports 8 bit
    RGB imagery and offers experimental support for 16 bit RGB imagery.
    Does not support Multispectral imagery yet.
    Implementation based on https://doi.org/10.1109/CVPR.2019.01063 .

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            ``prepare_data`` function.
    ---------------------   -------------------------------------------
    backbone                Optional String. Backbone CNN model to be used for
                            creating the base. If hourglass is chosen as
                            the mtl_model (Architecture),then this parameter is
                            ignored as hourglass uses a special customised
                            architecture.
                            This parameter is used with `linknet` model.
                            Default: 'resnet34'

                            Supported backbones: 'resnet18', 'resnet34',
                            'resnet50', 'resnet101', 'resnet152'
    ---------------------   -------------------------------------------
    mtl_model               Optional String. It is used to create model
                            from linknet or hourglass based neural architectures.
                            Supported: 'linknet', 'hourglass'.
                            Default: 'hourglass'
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where a compatible pre-trained
                            model is saved. Accepts a Deep Learning Package
                            (DLPK) or Esri Model Definition(EMD) file.
    =====================   ===========================================

    **kwargs**

    =============================   =============================================
    **Argument**                    **Description**
    -----------------------------   ---------------------------------------------
    gaussian_thresh                 Optional float. Sets the gaussian threshold
                                    which allows to set the required road width.
                                    Range: 0.0 to 1.0
                                    Default:0.76
    -----------------------------   ---------------------------------------------
    orient_bin_size                 Optional Int. Sets the bin size for
                                    orientation angles.
                                    Default:20
    -----------------------------   ---------------------------------------------
    orient_theta                    Optional Int. Sets the width of orientation
                                    mask.
                                    Default:8
    =============================   =============================================

    :returns: `MultiTaskRoadExtractor` Object
    """

    def __init__(
        self, data, backbone=None, pretrained_path=None, *args, **kwargs,
    ):
        #if data.sub_dataset_type != "RoadOrientation":
        #    raise Exception(
        #        "This model only works for road extraction. And In order to use this model it's corresponding model-specific parameters should also be passed at prepare_data."
        #    )
        # Set default backbone to be 'resnet34'
        self._validate_kwargs(**kwargs)
        if backbone is None:
            backbone = models.resnet34
        super().__init__(data, backbone, **kwargs)
        self._slice_lr = False  # Road models just have a single layer group due to which we cant slice the lr.
        predefined_mtl_model = None
        # Causes Divide by zero error in  fastai library
        if hasattr(self._data, "orig_path"):  # If true the data is not empty class obtained from from_model
            if pretrained_path is not None:
                if os.path.isdir(pretrained_path):
                    for file in os.listdir(pretrained_path):
                        if file.endswith(".emd"):
                            try:
                                emd = json.load(open(os.path.join(pretrained_path, file)))
                                bin_size = emd['RoadOrientation']['orient_bin_size']
                                predefined_mtl_model = emd['ModelParameters']['mtl_model']
                            except Exception as e:
                                raise e
                        elif file.endswith(".dlpk"):
                            try:
                                emd_path = _get_emd_path(Path(file))
                                with open(emd_path) as f:
                                    emd = json.load(f)
                                bin_size = emd['RoadOrientation']['orient_bin_size']
                                predefined_mtl_model = emd['ModelParameters']['mtl_model']
                            except Exception as e:
                                raise e
                else:
                    try:
                        emd_path = _get_emd_path(pretrained_path)
                        with open(emd_path) as f:
                            emd = json.load(f)
                        bin_size = emd['RoadOrientation']['orient_bin_size']
                        predefined_mtl_model = emd['ModelParameters']['mtl_model']
                    except Exception as e:
                        raise e
                try:
                    bin_size
                except:
                    print('Could not find the emd file in the specified path. Please provide correct path')

                kwargs['orient_bin_size'] = bin_size #To ensure the data is consistent across multiple runs

            self._orient_data=self._get_road_orient_data(data,**kwargs)
            if len(data.classes)>2:
                raise Exception(
                    "Found multi-labels in the data, This is a binary segmentation model and hence please export the data with binary labels."  # noqa
                )
            self._data.classes = self._orient_data.classes
            class_key = self._orient_data.classes[0] if isinstance(self._orient_data.classes[0], int) else \
            self._orient_data.classes[1]
            self._data.class_mapping = {class_key: str(class_key)}
            self._orient_data.class_mapping = self._data.class_mapping
            self._orient_data._imagery_type = self._data._imagery_type if hasattr(data, '_imagery_type') else None
            if (self._data._imagery_type == 'MS'):
                self._orient_data._image_space_used = self._data._image_space_used if hasattr(data,
                                                                                              '_image_space_used') else None
                self._orient_data._is_multispectral = self._data._is_multispectral if hasattr(data,
                                                                                              '_is_multispectral') else None
                self._orient_data._band_max_values = self._data._band_max_values if hasattr(data,
                                                                                            '_band_max_values') else None
                self._orient_data._band_mean_values = self._data._band_mean_values if hasattr(data,
                                                                                              '_band_mean_values') else None
                self._orient_data._band_min_values = self._data._band_min_values if hasattr(data,
                                                                                            '_band_min_values') else None
                self._orient_data._band_std_values = self._data._band_std_values if hasattr(data,
                                                                                            '_band_std_values') else None
                self._orient_data._bands = self._data._bands if hasattr(data, '_bands') else None
                self._orient_data._extract_bands = self._data._extract_bands if hasattr(data,
                                                                                        '_extract_bands') else None
                self._orient_data._image_space_used = self._data._image_space_used if hasattr(data,
                                                                                              '_image_space_used') else None
                self._orient_data._min_max_scaler = self._data._min_max_scaler if hasattr(data,
                                                                                          '_min_max_scaler') else None
                self._orient_data._min_max_scaler_tfm = self._data._min_max_scaler_tfm if hasattr(data,
                                                                                                  '_min_max_scaler_tfm') else None
                self._orient_data._multispectral_color_array = self._data._multispectral_color_array if hasattr(data,
                                                                                                                '_multispectral_color_array') else None
                self._orient_data._multispectral_color_mapping = self._data._multispectral_color_mapping if hasattr(
                    data, '_multispectral_color_mapping') else None
                self._orient_data._norm_pct = self._data._norm_pct if hasattr(data, '_norm_pct') else None
                self._orient_data._rgb_bands = self._data._rgb_bands if hasattr(data, '_rgb_bands') else None
                self._orient_data._scaled_max_values = self._data._scaled_max_values if hasattr(data,
                                                                                                '_scaled_max_values') else None
                self._orient_data._scaled_mean_values = self._data._scaled_mean_values if hasattr(data,
                                                                                                  '_scaled_mean_values') else None
                self._orient_data._scaled_min_values = self._data._scaled_min_values if hasattr(data,
                                                                                                '_scaled_min_values') else None
                self._orient_data._scaled_std_values = self._data._scaled_std_values if hasattr(data,
                                                                                                '_scaled_std_values') else None
                self._orient_data._symbology_rgb_bands = self._data._symbology_rgb_bands if hasattr(data,
                                                                                                    '_symbology_rgb_bands') else None
                self._orient_data.stats = self._data.stats if hasattr(data, 'stats') else None
                self._orient_data._do_normalize = self._data._do_normalize if hasattr(data, '_do_normalize') else None
                self._orient_data.train_ds.tfms = self._data.train_ds.tfms if hasattr(data.train_ds, 'tfms') else None
                self._orient_data.valid_ds.tfms = self._data.valid_ds.tfms if hasattr(data.valid_ds, 'tfms') else None
        else:
            self._orient_data =self._data # Else use the data attributes obtained from emd
            #if self._orient_data._do_normalize:
            #    self._orient_data = self._orient_data.normalize(stats=(self._orient_data._scaled_mean_values, self._orient_data._scaled_std_values), do_x=True, do_y=False)
        #self._orient_data.train_ds.x._imagery_type = self._orient_data._imagery_type
        #self._orient_data.valid_ds.x._imagery_type = self._orient_data._imagery_type
        self._ignore_classes = kwargs.get("ignore_classes", [])
        if self._ignore_classes != [] and len(self._orient_data.classes) <= 3:
            raise Exception(
                "`ignore_classes` parameter can only be used when the dataset has more than 2 classes."  # noqa
            )

        data_classes = list(self._orient_data.class_mapping.keys())
        if 0 not in list(self._orient_data.class_mapping.values()):
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes if k != 0
            ]
        else:
            self._ignore_mapped_class = [
                data_classes.index(k) + 1 for k in self._ignore_classes
            ]
        if self._ignore_classes != []:
            if 0 not in self._ignore_mapped_class:
                self._ignore_mapped_class.insert(0, 0)

        pixel_accuracy = partial(
            accuracy, ignore_mapped_class=self._ignore_mapped_class
        )
        road_iou = partial(miou, ignore_mapped_class=self._ignore_mapped_class)
        dice_coeff = partial(dice, ignore_classes=[[0]])

        n_bands = len(getattr(self._data, "_extract_bands", [0, 1, 2]))
        _backbone = self._backbone
        if hasattr(self, "_orig_backbone"):
            _backbone = self._orig_backbone

        # Check if a backbone provided is compatible, use resnet50 as default
        if not self._check_backbone_support(_backbone):
            raise Exception(
                f"Enter only compatible backbones from {', '.join(self.supported_backbones)}"  # noqa
            )

        self.name = "MultiTaskRoadExtractor"
        self._code = image_classifier_prf

        self._chip_size = (self._orient_data.chip_size, self._orient_data.chip_size)

        # Cut-off the backbone before the penultimate layer
        self._encoder = create_body(self._backbone, -2)

        # Initialize the model, loss function and the Learner object
        mtl_models = {
            "linknet": LinkNetMultiTaskModel,
            "hourglass": StackHourglassMultiTaskModel,
        }
        if predefined_mtl_model:
            self._mtl_model = predefined_mtl_model
        else:
            self._mtl_model = kwargs.get("mtl_model", "hourglass")
            self._mtl_model = (
                self._mtl_model if self._mtl_model in mtl_models.keys() else "hourglass"
            )
        self._model_init_kwargs = kwargs.get("model_init_kwargs", {})
        self._model = mtl_models[self._mtl_model](
            self._encoder,
            task1_classes=self._orient_data.c,
            task2_classes=self._orient_data.orient_c,
            chip_size=self._chip_size,
            n_bands=n_bands,
            **self._model_init_kwargs,
        )
        self._loss_f = road_orient_loss(
            n_classes=self._orient_data.c, loss_weights=kwargs.get("loss_weights", (1.0, 1.0))
        )
        self.learner_params = kwargs.get("learner_params", {})
        learner_kwargs = self.learner_params.copy()
        self._opt_func_name = kwargs.get("opt_func", None)
        if self._opt_func_name:
            _opt_func = self._get_optimizer(self._opt_func_name)
            if _opt_func:
                self._opt_func_args = kwargs.get("opt_func_args", {})
                learner_kwargs["opt_func"] = partial(_opt_func, **self._opt_func_args)

        self.learn = MultiTaskRoadLearner(
            self._orient_data,
            self._model,
            loss_func=self._loss_f,
            metrics=[pixel_accuracy, road_iou, dice_coeff],
            **learner_kwargs,
        )
        self.learn.path = self._data.path
        self.learn.model = self.learn.model.to(self._device)
        if pretrained_path is not None:
            super().load(str(pretrained_path))
        self._arcgis_init_callback()  # make first conv weights learnable

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "<%s>" % (type(self).__name__)

    def _get_road_orient_data(self,data,**kwargs):
        #if (not self._data._imagery_type in ['RGB', 'ASSUMED_RGB']):
        #    # TODO: Implement Road Orientation model for MS imagery
        #    raise NotImplementedError('Road Orientation Model does not suppport Multispectral imagery yet!')
        road_data_obj = ClassifiedTilesData(
            path=self._data.orig_path, class_mapping={}, chip_size=self._data.chip_size, val_split_pct=self._data._val_split_pct,
            batch_size=self._data.batch_size, transforms=None, seed=42,
            dataset_type=self._data.dataset_type, resize_to=self._data.resize_to, **kwargs)
        road_orient_obj = RoadOrientation(road_data_obj, **kwargs)
        orient_data = road_orient_obj.get_databunch(data,**kwargs)
        return orient_data

    def _get_optimizer(self, optim_name):
        optim_funcs = {"sgd": optim.SGD}
        return optim_funcs.get(optim_name, None)

    def _validate_kwargs(self, **kwargs):
        gauss_thresh = kwargs.get("gaussian_thresh", 0.6)
        orient_bin_size = kwargs.get("orient_bin_size", 20)
        orient_theta = kwargs.get("orient_theta", 8)
        assert gauss_thresh > 0.0 and gauss_thresh < 1, 'gauss_thresh should be between 0 and 1'
        assert orient_bin_size > 0 and orient_bin_size < 360, 'orient_bin_size should be between 1 and 360'
        assert orient_theta > 0 and orient_theta < 10, 'orient_theta should be between 1 and 10'

    def unfreeze(self):
        """
        Unfreezes the earlier layers of the model for fine-tuning.
        """
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    # Return a list of supported backbones names
    @property
    def supported_backbones(self):
        """
        Supported torchvision backbones for this model.
        """
        return MultiTaskRoadExtractor._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    @property
    def supported_datasets(self):
        """ Supported dataset types for this model. """
        return MultiTaskRoadExtractor._supported_datasets()

    @staticmethod
    def _supported_datasets():
        return ['Classified_Tiles']

    def fit(self, epochs=10, lr=None, **kwargs):
        save_callback_params = {
            "monitor": "miou",
            "every": "improvement",
            "load_best_at_end": False,
        }
        save_callback_params.update(kwargs.get("save_callback_params", {}))
        kwargs.update(save_callback_params=save_callback_params)
        if isinstance(lr, slice):
            lr = lr.stop
        super().fit(epochs, lr=lr, **kwargs)

    def _get_emd_params(self,save_inference_file):
        _emd_template = {}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISImageClassifier.py"
        else:
            _emd_template[
                "InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISImageClassifier.py"
        _emd_template["ModelConfiguration"] = "_road_infrencing"
        _emd_template["ModelType"] = "ImageClassification"
        _emd_template["ExtractBands"] = [0, 1, 2]

        model_params = {
            "backbone": self._backbone,
            "backend": self._backend,
            "opt_func": self._opt_func_name if hasattr(self, "_opt_func_name") else "",
            "opt_func_args": self._opt_func_args
            if hasattr(self, "_opt_func_args")
            else "",
            "loss": self._loss_f.__class__.__name__,
            "loss_weights": self._loss_f.__dict__["loss_weights"],
            "mtl_model": self._mtl_model,
            "model_init_kwargs": self._model_init_kwargs,
            "learner_params": self.learner_params,
        }
        _emd_template["ModelParameters"] = model_params
        _emd_template["Classes"] = []

        class_data = {}
        # 0th index is background
        for _, class_name in enumerate(self._data.classes[1:]):
            inverse_class_mapping = {
                int(v): k for k, v in self._data.class_mapping.items()
            }
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)]
            class_data["Color"] = color
            _emd_template["Classes"].append(class_data.copy())

        if hasattr(self._orient_data, "parent_obj"):
            data_params = {}
            for param_key, param_value in self._orient_data.parent_obj.__dict__.items():
                if "files" in param_key:
                    continue
                if hasattr(param_value, "__dict__"):
                    continue
                if not safe_json(param_value):
                    param_value = str(param_value)
                data_params[f"{param_key}"] = param_value
            _emd_template[self._orient_data.parent_obj.__class__.__name__] = data_params

        return _emd_template

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates a Multi-Task Learning model for binary segmentation from a
        Deep Learning Package(DLPK) or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_data` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `Multi-Task Road Extractor` Object
        """
        if not HAS_FASTAI:
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)
        model_file = Path(emd["ModelFile"])
        chip_size = emd["ImageWidth"]
        model_params = emd["ModelParameters"]
        orient_c = int(360.0 / emd["RoadOrientation"]["orient_bin_size"]) + 1

        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file

        class_mapping = {i["Value"]: i["Name"] for i in emd["Classes"]}

        if data is None:
            force_cpu = _device_check()
            data = _EmptyData(
                path=emd_path.parent.parent,
                loss_func=None,
                c=len(class_mapping) + 1,
                chip_size=chip_size,
            )
            data.orient_c = orient_c
            data.class_mapping = class_mapping
            data = get_multispectral_data_params_from_emd(data, emd)
            data.emd_path = emd_path
            data.emd = emd

        return cls(data, **model_params, pretrained_path=emd_path)

    def show_results(self, rows=2, **kwargs):
        """
        Shows the ground truth and predictions of model side by side.

        **kwargs**

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Number of rows of data to be displayed, if
                                batch size is smaller, then the rows will
                                display the value provided for batch size.
        ---------------------   -------------------------------------------
        alpha                   Optional Float. Opacity parameter for label
                                overlay on image. Float [0.0 - 1.0]
                                Default: 0.6
        =====================   ===========================================

        """
        self._check_requisites()
        self.return_fig = kwargs.get("return_fig", False)
        fig=self.learn.show_results(rows=rows, **kwargs)
        if self.return_fig:
            return fig

    @property
    def _model_metrics(self):
        acc, mean_iou = self._get_model_metrics()
        return {
            "accuracy": "{}".format(acc),
            "mIoU": "{}".format(mean_iou),
        }

    def _get_model_metrics(self, **kwargs):
        checkpoint = kwargs.get("checkpoint", True)
        if not hasattr(self.learn, "recorder"):
            return 0.0, 0.0

        try:
            model_accuracy = self.learn.recorder.metrics[-1][0]
            model_iou = self.learn.recorder.metrics[-1][1]
            if checkpoint:
                metrics_array = np.array(self.learn.recorder.metrics)
                model_iou = np.max(metrics_array[:, 1])
                model_accuracy = metrics_array[np.argmax(metrics_array[:, 1]), 0]
        except BaseException:
            logger = logging.getLogger()
            logger.debug("Cannot retrieve model accuracy.")
            model_accuracy = 0.0

        return float(model_accuracy), float(model_iou)



#!/usr/bin/env python3


"""
The methods are taken from => https://github.com/yxdragon/sknw

License:
==========================================================================
BSD 3-Clause License

Copyright (c) 2017, Yan xiaolong
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import networkx as nx
import numpy as np
from numba import jit


# get neighbors d index
def neighbors(shape):
    dim = len(shape)
    block = np.ones([3] * dim)
    block[tuple([1] * dim)] = 0
    idx = np.where(block > 0)
    idx = np.array(idx, dtype=np.uint8).T
    idx = np.array(idx - [1] * dim)
    acc = np.cumprod((1,) + shape[::-1][:-1])
    return np.dot(idx, acc[::-1])


@jit(nopython=True)  # my mark
def mark(img, nbs):  # mark the array use (0, 1, 2)
    img = img.ravel()
    for p in range(len(img)):
        if img[p] == 0:
            continue
        s = 0
        for dp in nbs:
            if img[p + dp] != 0:
                s += 1
        if s == 2:
            img[p] = 1
        else:
            img[p] = 2


@jit(nopython=True)  # trans index to r, c...
def idx2rc(idx, acc):
    rst = np.zeros((len(idx), len(acc)), dtype=np.int16)
    for i in range(len(idx)):
        for j in range(len(acc)):
            rst[i, j] = idx[i] // acc[j]
            idx[i] -= rst[i, j] * acc[j]
    rst -= 1
    return rst


@jit(nopython=True)  # fill a node (may be two or more points)
def fill(img, p, num, nbs, acc, buf):
    back = img[p]
    img[p] = num
    buf[0] = p
    cur = 0
    s = 1

    while True:
        p = buf[cur]
        for dp in nbs:
            cp = p + dp
            if img[cp] == back:
                img[cp] = num
                buf[s] = cp
                s += 1
        cur += 1
        if cur == s:
            break
    return idx2rc(buf[:s], acc)


# trace the edge and use a buffer, then buf.copy, if use [] numba not works
@jit(nopython=True)
def trace(img, p, nbs, acc, buf):
    c1 = 0
    c2 = 0
    newp = 0
    cur = 1
    while True:
        buf[cur] = p
        img[p] = 0
        cur += 1
        for dp in nbs:
            cp = p + dp
            if img[cp] >= 10:
                if c1 == 0:
                    c1 = img[cp]
                    buf[0] = cp
                else:
                    c2 = img[cp]
                    buf[cur] = cp
            if img[cp] == 1:
                newp = cp
        p = newp
        if c2 != 0:
            break
    return (c1 - 10, c2 - 10, idx2rc(buf[:cur + 1], acc))


@jit(nopython=True)  # parse the image then get the nodes and edges
def parse_struc(img, pts, nbs, acc):
    img = img.ravel()
    buf = np.zeros(131072, dtype=np.int64)
    num = 10
    nodes = []
    for p in pts:
        if img[p] == 2:
            nds = fill(img, p, num, nbs, acc, buf)
            num += 1
            nodes.append(nds)
    edges = []
    for p in pts:
        for dp in nbs:
            if img[p + dp] == 1:
                edge = trace(img, p + dp, nbs, acc, buf)
                edges.append(edge)
    return nodes, edges

# use nodes and edges build a networkx graph


def build_graph(nodes, edges, multi=False):
    graph = nx.MultiGraph() if multi else nx.Graph()
    for i in range(len(nodes)):
        graph.add_node(i, pts=nodes[i], o=nodes[i].mean(axis=0))
    for s, e, pts in edges:
        ln = np.linalg.norm(pts[1:] - pts[:-1], axis=1).sum()
        graph.add_edge(s, e, pts=pts, weight=ln)
    return graph


def buffer(ske):
    buf = np.zeros(tuple(np.array(ske.shape) + 2), dtype=np.uint16)
    buf[tuple([slice(1, -1)] * buf.ndim)] = ske
    return buf


def mark_node(ske):
    buf = buffer(ske)
    nbs = neighbors(buf.shape)
    acc = np.cumprod((1,)+buf.shape[::-1][:-1])[::-1]
    mark(buf, nbs)
    return buf


def build_sknw(ske, multi=False):
    buf = buffer(ske)
    nbs = neighbors(buf.shape)
    acc = np.cumprod((1,)+buf.shape[::-1][:-1])[::-1]
    mark(buf, nbs)
    pts = np.array(np.where(buf.ravel() == 2))[0]
    nodes, edges = parse_struc(buf, pts, nbs, acc)
    return build_graph(nodes, edges, multi)

# draw the graph


def draw_graph(img, graph, cn=255, ce=128):
    acc = np.cumprod((1,) + img.shape[::-1][:-1])[::-1]
    img = img.ravel()
    for (s, e) in graph.edges():
        eds = graph[s][e]
        if isinstance(graph, nx.MultiGraph):
            for i in eds:
                pts = eds[i]['pts']
                img[np.dot(pts, acc)] = ce
        else:
            img[np.dot(eds['pts'], acc)] = ce
    for idx in graph.nodes():
        pts = graph.nodes[idx]['pts']
        img[np.dot(pts, acc)] = cn


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    img = np.array([
        [0, 0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0, 0],
        [1, 1, 1, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 1, 1, 1, 1],
        [0, 0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0, 0]])

    node_img = mark_node(img)
    graph = build_sknw(img)

    plt.imshow(node_img[1:-1, 1:-1], cmap='gray')

    # draw edges by pts
    for (s, e) in graph.edges():
        ps = graph[s][e]['pts']
        plt.plot(ps[:, 1], ps[:, 0], 'green')

    # draw node by o
    nodes = graph.nodes()
    ps = np.array([nodes[i]['o'] for i in nodes])
    plt.plot(ps[:, 1], ps[:, 0], 'r.')

    # title and show
    plt.title('Build Graph')
    plt.show()



""" Defines store functions for working with AssignmentTypes in Version 2 projects.
"""

from ... import workforce
from .utils import validate, add_features, update_features, remove_features


def get_assignment_type_v2(project, code=None, name=None):
    """ Gets the identified AssignmentType. Exactly one form of identification should be provided.
        :param project:
        :param code: The AssignmentType GlobalID.
        :param name: The AssignmentType name.
    """
    assignment_types = get_assignment_types_v2(project)
    if code:
        return next((at for at in assignment_types if at.code == code), None)
    elif name:
        return next((at for at in assignment_types if at.name == name), None)
    return None


def get_assignment_types_v2(project):
    """ Gets all AssignmentTypes for the project.
        :param project:
        :returns: A list of AssignmentTypes.
    """
    return query_assignment_types(project, '1=1')


def query_assignment_types(project, where='1=1'):
    """ Executes a query against the assignment types table.
        :param project: The project in which to query assignment types.
        :param where: An ArcGIS where clause.
        :returns: list of Assignment Types
    """
    assignment_type_features = project.assignment_types_table.query(where, return_all_records=True).features
    return [workforce.AssignmentType(project, feature) for feature in assignment_type_features]


def add_assignment_type_v2(project, name):
    """
    Adds a new assignment type
    """
    assignment_type = workforce.AssignmentType(project, name=name)
    return add_assignment_types_v2(project, [assignment_type])[0]


def add_assignment_types_v2(project, assignment_types):
    """ Adds an AssignmentType to a project.
        :param project:
        :param assignment_types: list of AssignmentTypes
        :raises ValidationError: Indicates that one or more assignment types failed validation.
    """
    for assignment_type in assignment_types:
        validate(assignment_type._validate)

    use_global_ids = False
    features = [assignment_type.feature for assignment_type in assignment_types]
    add_features(project.assignment_types_table, features, use_global_ids)
    project._update_cached_assignment_types()
    return assignment_types


def update_assignment_types_v2(project, assignment_types):
    """ Updates the AssignmentTypes.
        :param project:
        :param assignment_types: list of AssignmentTypes
        :raises ValidationError: Indicates that one or more assignment types failed validation.
    """
    for assignment_type in assignment_types:
        validate(assignment_type._validate_for_update)
    features = [assignment_type.feature for assignment_type in assignment_types]
    update_features(project.assignment_types_table, features)
    project._update_cached_assignment_types()
    return assignment_types


def update_assignment_type_v2(project, assignment_type, name=None):
    """Updates an assignment type's name"""
    if name:
        assignment_type.name = name
    return update_assignment_types_v2(project, [assignment_type])[0]


def delete_assignment_types_v2(project, assignment_types):
    """ Removes AssignmentTypes from the project.
        :param project:
        :param assignment_types: list of AssignmentTypes.
    """
    for assignment_type in assignment_types:
        validate(assignment_type._validate_for_remove)
    features = [assignment_type.feature for assignment_type in assignment_types]
    remove_features(project.assignment_types_table, features)
    project._update_cached_assignment_types()



import torch
from torch import nn, LongTensor
import torch.nn.functional as F
from fastai.vision import imagenet_stats
from fastai.vision.image import ImageBBox
from fastai.vision.data import ObjectCategoryList, ObjectItemList
from fastai.core import split_kwargs_by_func
from fastprogress.fastprogress import progress_bar

import numpy as np
import random
import math

def conv_params(in_size, out_size):
    filters = [3,2,5,4]
    strides = [1,2,3] # max_stride = 3
    pads = [0,1,2,3] # max pad
    
    if out_size == 1:
        return 1, 0, in_size
    
    for filter_size in filters:
        for pad in pads:
            for stride in strides:
                if ((out_size - 1) * stride == (in_size - filter_size) + 2 * pad):
                    return stride, pad, filter_size
    return None, None, None

def conv_paramsv2(in_size, out_size):
    filters = [3]
    strides = [1,2] # max_stride = 2
    pads = [0,1] # max pad
    
    if out_size == 1:
        return 1, 0, in_size
    
    for filter_size in filters:
        for pad in pads:
            for stride in strides:
                if (((in_size - filter_size) + 2 * pad) // stride) + 1 == out_size:
                    return stride, pad, filter_size
    return None, None, None


class StdConv(nn.Module):
    def __init__(self, nin, nout, filter_size=3, stride=2, padding=1, drop=0.1):
        super().__init__()
        self.conv = nn.Conv2d(nin, nout, filter_size, stride=stride, padding=padding)
        self.bn = nn.BatchNorm2d(nout)
        self.drop = nn.Dropout(drop)
        
    def forward(self, x): 
        return self.drop(self.bn(F.relu(self.conv(x))))

class StdConvv2(nn.Module):
    def __init__(self, nin, nout, upsample_size=0, upsample=False, filter_size=3, stride=2, padding=1, drop=0.1):
        super().__init__()
        self.upsample = upsample
        self.conv = nn.Conv2d(nin, nout, filter_size, stride=stride, padding=padding)
        self.up = nn.Upsample(upsample_size)
        self.bn = nn.BatchNorm2d(nout)
        self.drop = nn.Dropout(drop)
        
    def forward(self, x):
        if self.upsample == True:
             return self.drop(self.bn(F.relu(self.conv(self.up(x)))))
        return self.drop(self.bn(F.relu(self.conv(x))))
            
def flatten_conv(x,k):
    bs,nf,gx,gy = x.size()
    x = x.permute(0,2,3,1).contiguous()
    return x.view(bs,-1,nf//k)

class OutConv(nn.Module):
    def __init__(self, k, nin, num_classes, bias):
        super().__init__()
        self.k = k
        self.oconv1 = nn.Conv2d(nin, (num_classes)*k, 3, padding=1)
        self.oconv2 = nn.Conv2d(nin, 4*k, 3, padding=1)
        self.oconv1.bias.data.zero_().add_(bias)
        
    def forward(self, x):
        return [flatten_conv(self.oconv1(x), self.k),
                flatten_conv(self.oconv2(x), self.k)]
    
class SSDHead(nn.Module):
    def __init__(self, grids, anchors_per_cell, num_classes, num_features=7, drop=0.3, bias=-4., num_channels=512):
        super().__init__()
        self.drop = nn.Dropout(drop)
        
        self.sconvs = nn.ModuleList([])
        self.oconvs = nn.ModuleList([])
        
        self.anc_grids = grids
        
        self._k = anchors_per_cell

        
        self.sconvs.append(StdConv(num_channels, 256, stride=1, drop=drop))
        
        
        for i in range(len(grids)):
            
            if i == 0:
                stride, pad, filter_size = conv_params(num_features, grids[i])
            else:
                stride, pad, filter_size = conv_params(grids[i-1], grids[i])
            
            if stride is None:
                print(grids[i-1], ' --> ', grids[i])
                raise Exception('cannot create model for specified grids.')
                
            self.sconvs.append(StdConv(256, 256, filter_size, stride=stride, padding=pad, drop=drop))
            self.oconvs.append(OutConv(self._k, 256, num_classes=num_classes, bias=bias))
       
    def forward(self, x):
        x = self.drop(F.relu(x))
        x = self.sconvs[0](x)
        out_classes = []
        out_bboxes = []
        for sconv, oconv in zip(self.sconvs[1:], self.oconvs):
            x = sconv(x)
            out_class, out_bbox = oconv(x)
            out_classes.append(out_class)
            out_bboxes.append(out_bbox)
            
        return [torch.cat(out_classes, dim=1),
                torch.cat(out_bboxes, dim=1)]

class SSDHeadv2(nn.Module):
    def __init__(self, grids, anchors_per_cell, num_classes, num_features=7, drop=0.3, bias=-4., num_channels=512):
        super().__init__()
        self.drop = nn.Dropout(drop)
        
        self.sconvs = nn.ModuleList([])
        self.oconvs = nn.ModuleList([])
        
        self.anc_grids = grids
        
        self._k = anchors_per_cell

        
        self.sconvs.append(StdConvv2(num_channels, 256, stride=1, drop=drop))
        
        
        for i in range(len(grids)):
            
            upsample = False

            if i == 0 and num_features >= grids[i]:
                stride, pad, filter_size = conv_paramsv2(num_features, grids[i])
                if stride is None:
                    upsample = True
                    stride, pad, filter_size = 1, 1, 3
            elif i == 0 and num_features < grids[i]:
                upsample = True
                stride, pad, filter_size = 1, 1, 3

            elif i != 0 and grids[i-1] > grids[i]:
                stride, pad, filter_size = conv_paramsv2(grids[i-1], grids[i])
                if stride is None:
                    upsample=True
                    stride, pad, filter_size = 1,1,3
            else:
                upsample=True
                stride, pad, filter_size = 1,1,3 
                
            self.sconvs.append(StdConvv2(256, 256, grids[i], upsample, filter_size, stride=stride, padding=pad, drop=drop))
            self.oconvs.append(OutConv(self._k, 256, num_classes=num_classes, bias=bias))
                
    def forward(self, x):
        x = self.drop(F.relu(x))
        x = self.sconvs[0](x)
        out_classes = []
        out_bboxes = []
        for sconv, oconv in zip(self.sconvs[1:], self.oconvs):
            x = sconv(x)
            out_class, out_bbox = oconv(x)
            out_classes.append(out_class)
            out_bboxes.append(out_bbox)
            
        return [torch.cat(out_classes, dim=1),
                torch.cat(out_bboxes, dim=1)]


def one_hot_embedding(labels, num_classes):
    return torch.eye(num_classes)[labels.data.cpu()]

class BCE_Loss(nn.Module):
    def __init__(self, num_classes):
        super().__init__()
        self.num_classes = num_classes

    def forward(self, pred, targ):
        t = one_hot_embedding(targ, self.num_classes)
        t = torch.Tensor(t[:,1:].contiguous()).to(pred.device)
        x = pred[:,1:]
        w = self.get_weight(x,t)
        return F.binary_cross_entropy_with_logits(x, t, w, reduction='sum')/(self.num_classes-1)
    
    def get_weight(self,x,t): return None

class FocalLoss(BCE_Loss):
    def get_weight(self,x,t):
        alpha,gamma = 0.25,1
        p = x.sigmoid()
        pt = p*t + (1-p)*(1-t)
        w = alpha*t + (1-alpha)*(1-t)
        w = w * (1-pt).pow(gamma)
        return w.detach()

def nms(boxes, scores, overlap=0.5, top_k=100):
    keep = scores.new(scores.size(0)).zero_().long()
    if boxes.numel() == 0: return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    area = torch.mul(x2 - x1, y2 - y1)
    v, idx = scores.sort(0)  # sort in ascending order
    idx = idx[-top_k:]  # indices of the top-k largest vals
    xx1 = boxes.new()
    yy1 = boxes.new()
    xx2 = boxes.new()
    yy2 = boxes.new()
    w = boxes.new()
    h = boxes.new()

    count = 0
    while idx.numel() > 0:
        i = idx[-1]  # index of current largest val
        keep[count] = i
        count += 1
        if idx.size(0) == 1: break
        idx = idx[:-1]  # remove kept element from view
        # load bboxes of next highest vals
        torch.index_select(x1, 0, idx, out=xx1)
        torch.index_select(y1, 0, idx, out=yy1)
        torch.index_select(x2, 0, idx, out=xx2)
        torch.index_select(y2, 0, idx, out=yy2)
        # store element-wise max with next highest score
        xx1 = torch.clamp(xx1, min=x1[i])
        yy1 = torch.clamp(yy1, min=y1[i])
        xx2 = torch.clamp(xx2, max=x2[i])
        yy2 = torch.clamp(yy2, max=y2[i])
        w.resize_as_(xx2)
        h.resize_as_(yy2)
        w = xx2 - xx1
        h = yy2 - yy1
        # check sizes of xx1 and xx2.. after each iteration
        w = torch.clamp(w, min=0.0)
        h = torch.clamp(h, min=0.0)
        inter = w*h
        # IoU = i / (area(a) + area(b) - i)
        rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
        union = (rem_areas - inter) + area[i]
        IoU = inter/union  # store result in iou
        # keep only elements with an IoU <= overlap
        idx = idx[IoU.le(overlap)]
    return keep, count

def postprocess(pred, model=None, thresh=0.5, nms_overlap=0.1, ret_scores=True, device=None):
    """
    It works on a single activation, does not support batch.
    """
    if device is None:
        device = torch.device('cpu')
    b_clas, b_bb = pred
    a_ic = model._actn_to_bb(b_bb.to(device), model._anchors.to(device), model._grid_sizes.to(device))
    conf_scores, clas_ids = b_clas[:, 1:].max(1)
    conf_scores = b_clas.t().sigmoid().to(device)

    out1, bbox_list, class_list = [], [], []

    for cl in range(1, len(conf_scores)):
        c_mask = conf_scores[cl] > thresh
        if c_mask.sum() == 0: 
            continue
        scores = conf_scores[cl][c_mask]
        l_mask = c_mask.unsqueeze(1)
        l_mask = l_mask.expand_as(a_ic)
        boxes = a_ic[l_mask].view(-1, 4) # boxes are now in range[ 0, 1]
        boxes = (boxes-0.5) * 2.0        # putting boxes in range[-1, 1]
        ids, count = nms(boxes.data, scores, nms_overlap, 50) # FIX- NMS overlap hardcoded
        ids = ids[:count]
        out1.append(scores[ids])
        bbox_list.append(boxes.data[ids])
        class_list.append(torch.tensor([cl]*count))

    if len(bbox_list) == 0:
        return None

    if ret_scores:
        return torch.cat(bbox_list, dim=0).to(device), torch.cat(class_list, dim=0).to(device), torch.cat(out1, dim=0).to(device)
    else:
        return torch.cat(bbox_list, dim=0), torch.cat(class_list, dim=0)

def compute_ap(precision, recall):
    "Compute the average precision for `precision` and `recall` curve."
    recall = np.concatenate(([0.], list(recall), [1.]))
    precision = np.concatenate(([0.], list(precision), [0.]))
    for i in range(len(precision) - 1, 0, -1):
        precision[i - 1] = np.maximum(precision[i - 1], precision[i])
    idx = np.where(recall[1:] != recall[:-1])[0]
    ap = np.sum((recall[idx + 1] - recall[idx]) * precision[idx + 1])
    return ap

def compute_class_AP(model, dl, n_classes, show_progress, iou_thresh=0.5, detect_thresh=0.35, num_keep=100, **kwargs):
    tps, clas, p_scores = [], [], []
    if getattr(model, "_is_model_extension", False):
        transform_kwargs, kwargs = split_kwargs_by_func(kwargs, model.model_conf.transform_input)
    classes, n_gts = LongTensor(range(n_classes)),torch.zeros(n_classes).long()
    with torch.no_grad():
        for input,target in progress_bar(dl, display=show_progress):

            if getattr(model, "_is_model_extension", False):
                try:
                    if model._is_multispectral:
                        output = model.learn.model.eval()(model.model_conf.transform_input_multispectral(input, **transform_kwargs))
                    else:
                        output = model.learn.model.eval()(model.model_conf.transform_input(input, **transform_kwargs))
                except Exception as e:

                    if getattr(model, "_is_fasterrcnn", False):
                        output = []
                        for _ in range(input.shape[0]):
                            res={}
                            res['boxes'] = torch.empty(0,4)
                            res['scores'] = torch.tensor([])
                            res['labels'] = torch.tensor([])
                            output.append(res)
                    else:
                        raise e
                analyzed_pred_out = model._analyze_pred(output,
                                                        thresh=detect_thresh,
                                                        nms_overlap=iou_thresh,
                                                        ret_scores=True,
                                                        device=model._device,
                                                        )
            else:
                output = model.learn.pred_batch(batch=(input, target))
            for i in range(target[0].size(0)):
                if getattr(model, "_is_model_extension", False):
                    op = analyzed_pred_out[i]
                else:
                    op = model._data.y.analyze_pred((output[0][i], output[1][i]), model=model, thresh=detect_thresh, nms_overlap=iou_thresh, ret_scores=True, device=model._device)
                tgt_bbox, tgt_clas = model._get_y(target[0][i], target[1][i])
                
                try:
                    bbox_pred, preds, scores = op
                    if len(bbox_pred) != 0 and len(tgt_bbox) != 0:
                        ious = model._jaccard(bbox_pred, tgt_bbox)
                        max_iou, matches = ious.max(1)
                        detected = []
                        for i in range(len(preds)):
                            if max_iou[i] >= iou_thresh and matches[i] not in detected and tgt_clas[matches[i]] == preds[i]:
                                detected.append(matches[i])
                                tps.append(1)
                            else: tps.append(0)
                        clas.append(preds.cpu())
                        p_scores.append(scores.cpu())
                except Exception as e:
                    pass
                n_gts += ((tgt_clas.cpu()[:,None] - 1) == classes[None,:]).sum(0)

    # If no true positives are found return an average precision score of 0.
    if len(tps) == 0: return [0. for cls in range(1, n_classes + 1)]

    tps, p_scores, clas = torch.tensor(tps), torch.cat(p_scores,0), torch.cat(clas,0)
    fps = 1-tps
    idx = p_scores.argsort(descending=True)
    tps, fps, clas = tps[idx], fps[idx], clas[idx]
    aps = []
    for cls in range(1,n_classes+1):
        tps_cls, fps_cls = tps[clas==cls].float().cumsum(0), fps[clas==cls].float().cumsum(0)
        if tps_cls.numel() != 0 and tps_cls[-1] != 0:
            precision = tps_cls / (tps_cls + fps_cls + 1e-8)
            recall = tps_cls / (n_gts[cls - 1] + 1e-8)
            aps.append(compute_ap(precision, recall))
        else: aps.append(0.)
    return aps

def iou(ann, centroids): 
    
    similarities = []

    for centroid in centroids:

        inter = np.prod(np.minimum(ann, centroid))       
        union = np.prod(ann) + np.prod(centroid) - inter
        similarities.append(inter/union)

    return np.array(similarities)


def avg_iou(bboxes, centroids):
    
    sum = 0.
    
    for bbox in bboxes:
        sum += max(iou(bbox, centroids))
        
    return sum/bboxes.shape[0]


def kmeans(bboxes, num_anchor):
    # Method based on https://github.com/experiencor/keras-yolo3

    num_points, dim = bboxes.shape
    prev_centroids = np.ones(num_points)*(-1)
    indices = [random.randrange(num_points) for i in range(num_anchor)]
    centroids = bboxes[indices]
    
    while True:
        distances = []
        for bbox in bboxes:
            d = 1 - iou(bbox, centroids)
            distances.append(d)
        distances = np.array(distances)
        cur_centroids = np.argmin(distances, axis=1)
        
        if (prev_centroids == cur_centroids).all() :
            return centroids
        
        centroid_sums = np.zeros((num_points, dim), np.float) #num_points needs to be num_anchors
        for i in range(num_points):
            centroid_sums[cur_centroids[i]] += bboxes[i]
        for i in range(num_anchor):
            centroids[i] = centroid_sums[i]/(np.sum(cur_centroids == i) + 1e-6)
            
        prev_centroids = cur_centroids.copy()



try:
    import tensorboardX
    from fastai.basics import *
    from fastai.vision import Learner
    from fastai.callbacks.tensorboard import LearnerTensorboardWriter
    from ..models._unet_utils import ArcGISSegmentationItemList
    from fastai.vision.data import ImageList, ImageImageList
    from torch.utils.tensorboard import SummaryWriter
    from fastai.callbacks.tensorboard import *
    from fastai.core import split_kwargs_by_func
    from PIL import Image
    from torchvision.transforms import ToTensor
except:
    pass


class ArcGISTBCallback(LearnerTensorboardWriter, Learner, ImageImageList, ArcGISSegmentationItemList, ImageList):
    def __init__(self, learn, base_dir, name, arcgis_model):
        self._base_dir = base_dir
        self._name = name
        self._arcgis_model = arcgis_model
        self._current_epoch = 0
        self._current_run = name + str(self._current_epoch)
        super(ArcGISTBCallback, self).__init__(learn, base_dir, self._current_run)

    def on_train_begin(self, **kwargs: Any):
        pass
    # Override the on_train_begin method of the parent class as it causes graph related errors and warnings.

    def on_epoch_end(self, last_metrics: MetricsList, iteration: int, **kwargs) -> None:
        self._current_epoch = self._current_epoch + 1
        self._current_run = self._name + '-Epoch-' + str(self._current_epoch)
        log_dir = self._base_dir / self._current_run
        self.tbwriter = SummaryWriter(str(log_dir))
        "Callback function that writes epoch end appropriate data to Tensorboard."
        if not all(x is None for x in last_metrics):
            self._write_metrics(iteration=iteration, last_metrics=last_metrics)
        # self._write_embedding(iteration=iteration)
        self._write_figure(iteration=iteration)

    def _write_figure(self, iteration: int) -> None:
        tag = 'predictions vs. actuals'
        rows = 4
        thresh = 0.1
        nms_overlap = 0.1
        obj_det_models = ['FeatureClassifier', 'SingleShotDetector', 'RetinaNet']
        img_to_img_models = ['UnetClassifier', 'SuperResolution', 'PSPNetClassifier', 'DeepLab']
        other_models = ['ImageCaptioner','MaskRCNN','MultiTaskRoadExtractor']
        text_models = ['TextClassifier']
        if (type(self._arcgis_model).__name__) in obj_det_models:
            fig1 = self.show_results(rows=rows, thresh=thresh, nms_overlap=nms_overlap, model=self._arcgis_model)
        elif (type(self._arcgis_model).__name__) in img_to_img_models:
            fig1 = self.show_results(rows=rows)  # Segmentation
        elif (type(self._arcgis_model).__name__) in other_models:
            fig1 = self._arcgis_model.show_results(2, return_fig=True)
        elif (type(self._arcgis_model).__name__) == 'FasterRCNN':
            if self.data._is_multispectral:
                fig1 = self._arcgis_model._show_results_multispectral(return_fig=True)
            else:
                fig1 = self._show_results_modified(2, return_fig=True)
        elif (type(self._arcgis_model).__name__) == 'CycleGAN':
            self._arcgis_model.learn.model.arcgis_results = True
            fig1 = self.show_results(rows=rows)
            self._arcgis_model.learn.model.arcgis_results = False
        elif (type(self._arcgis_model).__name__) == 'Pix2Pix':
            self._arcgis_model.learn.model.arcgis_results = True
            fig1 = self.show_results(rows=rows)
            self._arcgis_model.learn.model.arcgis_results = False
        else:
            return
        # Removing the support for Point CNN due to plotly orca dependency
        #if (type(self._arcgis_model).__name__) == 'PointCNN':
        #    fig1.write_image("fig1_pcnn.jpeg")
        #    image = Image.open("fig1_pcnn.jpeg")
        #    image = ToTensor()(image)
        #    self.tbwriter.add_image(tag=tag, img_tensor=image, global_step=iteration)
        # elif (type(self._arcgis_model).__name__) in text_models:
        # self.tbwriter.add_text(tag=tag, text_string =txt, global_step=iteration)
        #else:
        self.tbwriter.add_figure(tag=tag, figure=fig1, global_step=iteration, close=True)

    def show_results(self, ds_type=DatasetType.Valid, rows: int = 5, **kwargs):
        "Show `rows` result of predictions on `ds_type` dataset."
        n_items = rows ** 2 if self.data.train_ds.x._square_show_res else rows
        if self.dl(ds_type).batch_size < n_items: n_items = self.dl(ds_type).batch_size
        ds = self.dl(ds_type).dataset
        self.callbacks.append(RecordOnCPU())
        preds = self.pred_batch(ds_type)
        *self.callbacks, rec_cpu = self.callbacks
        x, y = rec_cpu.input, rec_cpu.target
        norm = getattr(self.data, 'norm', False)
        if norm:
            x = self.data.denorm(x)
            if norm.keywords.get('do_y', False):
                y = self.data.denorm(y, do_x=True)
                preds = self.data.denorm(preds, do_x=True)
        analyze_kwargs, kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        if n_items > len(preds): n_items = len(preds)
        preds = [ds.y.analyze_pred(grab_idx(preds, i), **analyze_kwargs) for i in range(n_items)]
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(n_items)]
        if has_arg(ds.y.reconstruct, 'x'):
            ys = [ds.y.reconstruct(grab_idx(y, i), x=x) for i, x in enumerate(xs)]
            zs = [ds.y.reconstruct(z, x=x) for z, x in zip(preds, xs)]
        else:
            ys = [ds.y.reconstruct(grab_idx(y, i)) for i in range(n_items)]
            zs = [ds.y.reconstruct(z) for z in preds]
        if (type(self._arcgis_model).__name__) == 'FeatureClassifier' or (
                type(self._arcgis_model).__name__) == 'SingleShotDetector' or (
        type(self._arcgis_model).__name__) == 'RetinaNet':
            if self.data._is_multispectral:
                fig1 = self._arcgis_model._show_results_multispectral(return_fig=True)
            else:
                fig1 = self.show_xyzs(xs, ys, zs, **kwargs)
        elif (type(self._arcgis_model).__name__) == 'UnetClassifier' or (
        type(self._arcgis_model).__name__) == 'PSPNetClassifier' \
                or (type(self._arcgis_model).__name__) == 'DeepLab':
            if self.data._is_multispectral:
                fig1 = self._arcgis_model._show_results_multispectral(return_fig=True)
            else:
                fig1 = self.segment_show_xyzs(xs, ys, zs)
        elif (type(self._arcgis_model).__name__) == 'SuperResolution':
            fig1 = self.img_img_show_xyzs(xs, ys, zs)
        elif (type(self._arcgis_model).__name__) == 'CycleGAN':
            fig1 = self.img_tuple_show_xyzs(xs, ys, zs)
        elif (type(self._arcgis_model).__name__) == 'Pix2Pix':
            fig1 = self.img_tuple_show_xyzs(xs, ys, zs)
        return fig1

    def show_xyzs(self, xs, ys, zs, imgsize: int = 4, figsize: Optional[Tuple[int, int]] = None, **kwargs):
        "Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`."
        if (type(self._arcgis_model).__name__) == 'SingleShotDetector' or \
                (type(self._arcgis_model).__name__) == 'RetinaNet' or (
        type(self._arcgis_model).__name__) == 'FasterRCNN':
            self._square_show_res = False
        else:
            self._square_show_res = True
        if self._square_show_res:
            title = 'Ground truth\nPredictions'
            rows = int(np.ceil(math.sqrt(len(xs))))
            fig, axs = self.subplots(rows, rows, imgsize=imgsize, figsize=figsize, title=title, weight='bold', size=12)
            for x, y, z, ax in zip(xs, ys, zs, axs.flatten()):
                x.show(ax=ax, title=f'{str(y)}\n{str(z)}', **kwargs)
            for ax in axs.flatten()[len(xs):]: ax.axis('off')
        else:
            title = 'Ground truth/Predictions'
            fig, axs = self.subplots(len(xs), 2, imgsize=imgsize, figsize=figsize, title=title, weight='bold', size=14)
            for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
                x.show(ax=axs[i, 0], y=y, **kwargs)
                x.show(ax=axs[i, 1], y=z, **kwargs)

        return fig

    def show_xys(self, xs, ys, imgsize: int = 4, figsize: Optional[Tuple[int, int]] = None, **kwargs):
        "Show the `xs` (inputs) and `ys`(targets)  on a figure of `figsize`."
        fig, axs = self.subplots(len(xs), 2, imgsize=imgsize, figsize=figsize)
        for i, (x, y) in enumerate(zip(xs, ys)):
            x.show(ax=axs[i, 0], **kwargs)
            y.show(ax=axs[i, 1], **kwargs)
        plt.tight_layout()
        return fig

    def subplots(self, rows: int, cols: int, imgsize: int = 4, figsize: Optional[Tuple[int, int]] = None, title=None,
                 **kwargs):
        "Like `plt.subplots` but with consistent axs shape, `kwargs` passed to `fig.suptitle` with `title`"
        figsize = ifnone(figsize, (imgsize * cols, imgsize * rows))
        fig, axs = plt.subplots(rows, cols, figsize=figsize)
        if rows == cols == 1:
            axs = [[axs]]  # subplots(1,1) returns Axes, not [Axes]
        elif (rows == 1 and cols != 1) or (cols == 1 and rows != 1):
            axs = [axs]
        if title is not None: fig.suptitle(title, **kwargs)
        return fig, array(axs)

    def segment_show_xyzs(self, xs, ys, zs, imgsize: int = 4, figsize: Optional[Tuple[int, int]] = None, **kwargs):
        "Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`."
        if self._square_show_res:
            title = 'Ground truth\nPredictions'
            rows = int(np.ceil(math.sqrt(len(xs))))
            fig, axs = self.subplots(rows, rows, imgsize=imgsize, figsize=figsize, title=title, weight='bold', size=12)
            for x, y, z, ax in zip(xs, ys, zs, axs.flatten()): x.show(ax=ax, title=f'{str(y)}\n{str(z)}', **kwargs)
            for ax in axs.flatten()[len(xs):]: ax.axis('off')
        else:
            title = 'Ground truth/Predictions'
            fig, axs = self.subplots(len(xs), 2, imgsize=imgsize, figsize=figsize, title=title, weight='bold', size=14)
            for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
                x.show(ax=axs[i, 0], y=y, **kwargs)
                x.show(ax=axs[i, 1], y=z, **kwargs)
        return fig

    def img_img_show_xyzs(self, xs, ys, zs, imgsize: int = 4, figsize: Optional[Tuple[int, int]] = None, **kwargs):
        if (type(self._arcgis_model).__name__) == 'SuperResolution':
            self._square_show_res = False
            self._square_show = False
        "Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`."
        title = 'Input / Prediction / Target'
        fig, axs = self.subplots(len(xs), 3, imgsize=imgsize, figsize=figsize, title=title, weight='bold', size=14)
        for i, (x, y, z) in enumerate(zip(xs, ys, zs)):
            x.show(ax=axs[i, 0], **kwargs)
            y.show(ax=axs[i, 2], **kwargs)
            z.show(ax=axs[i, 1], **kwargs)
        return fig

    def img_tuple_show_xyzs(self, xs, ys, zs, figsize:Tuple[int,int]=None, **kwargs):
        """Show `xs` (inputs), `ys` (targets) and `zs` (predictions) on a figure of `figsize`.
        `kwargs` are passed to the show method."""
        figsize = ifnone(figsize, (12,3*len(xs)))
        fig,axs = plt.subplots(len(xs), 2, figsize=figsize)
        fig.suptitle('Ground truth / Predictions', weight='bold', size=14)
        for i,(x,z) in enumerate(zip(xs,zs)):
            x.to_one().show(ax=axs[i,0], **kwargs)
            z.to_one_pred().show(ax=axs[i,1], **kwargs)
        return fig

    def _show_results_modified(self, rows=5, **kwargs):

        if rows > len(self._arcgis_model._data.valid_ds):
            rows = len(self._arcgis_model._data.valid_ds)

        ds_type = DatasetType.Valid
        xb, yb = self._arcgis_model.learn.data.one_batch(ds_type, detach=False, denorm=False)
        ds = self._arcgis_model.learn.dl(ds_type).dataset
        if xb.shape[0]<rows ** 2:
            ds_type = DatasetType.Train
            xb, yb = self._arcgis_model.learn.data.one_batch(ds_type, detach=False, denorm=False)
            ds = self._arcgis_model.learn.dl(ds_type).dataset
        n_items = rows ** 2 if self._arcgis_model.learn.data.train_ds.x._square_show_res else rows
        if self._arcgis_model.learn.dl(ds_type).batch_size < n_items: n_items = self._arcgis_model.learn.dl(
            ds_type).batch_size
        self._arcgis_model.learn.model.eval()
        transform_kwargs, kwargs = split_kwargs_by_func(kwargs, self._arcgis_model.model_conf.transform_input)
        try:
            preds = self._arcgis_model.learn.model(self._arcgis_model.model_conf.transform_input(xb, transform_kwargs))
        except Exception as e:

            if getattr(self._arcgis_model, "_is_fasterrcnn", False):
                preds = []
                for _ in range(xb.shape[0]):
                    res={}
                    res['boxes'] = torch.empty(0,4)
                    res['scores'] = torch.tensor([])
                    res['labels'] = torch.tensor([])
                    preds.append(res)
            else:
                raise e

        x, y = to_cpu(xb), to_cpu(yb)
        norm = getattr(self._arcgis_model.learn.data, 'norm', False)
        if norm:
            x = self._arcgis_model.learn.data.denorm(x)
            if norm.keywords.get('do_y', False):
                y = self._arcgis_model.learn.data.denorm(y, do_x=True)
                preds = self._arcgis_model.learn.data.denorm(preds, do_x=True)
        analyze_kwargs, kwargs = split_kwargs_by_func(kwargs, ds.y.analyze_pred)
        preds = ds.y.analyze_pred(preds, self._arcgis_model, **analyze_kwargs)
        xs = [ds.x.reconstruct(grab_idx(x, i)) for i in range(n_items)]
        if has_arg(ds.y.reconstruct, 'x'):
            ys = [ds.y.reconstruct(grab_idx(y, i), x=x) for i, x in enumerate(xs)]
            zs = [ds.y.reconstruct(z, x=x) for z, x in zip(preds, xs)]
        else:
            ys = [ds.y.reconstruct(grab_idx(y, i)) for i in range(n_items)]
            zs = [ds.y.reconstruct(z) for z in preds]
        fig1 = self.show_xyzs(xs, ys, zs, **kwargs)
        return fig1



import uuid
from arcgis._impl.common._mixins import PropertyMap
###########################################################################
class BaseOGC(object):
    """

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required string. The web address of the endpoint.
    ---------------     --------------------------------------------------------------------
    gis                 Optional GIS. The `GIS` connection object
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    id                  Optional String. The unique ID of the layer.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and Layer List widgets.
    ===============     ====================================================================

    """
    _id = None
    _con = None
    _gis = None
    _url = None
    _type = "UnknownLayer"
    _title = None
    _opacity = None
    _copyright = None
    _min_scale = None
    _max_scale = None
    #----------------------------------------------------------------------
    def __init__(self, url, gis=None, **kwargs):
        self._url = url
        self._gis = gis
        self._min_scale, self._max_scale = kwargs.pop('scale', (0,0))
        self._title = kwargs.pop("title", "Layer")
        self._opacity = kwargs.pop('opacity', 1)
        self._id = kwargs.pop('id', uuid.uuid4().hex)
        self._copyright = kwargs.pop("copyright", "")
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """
        Returns the properties of the Layer.
        
        :returns: PropertyMap
        """
        return PropertyMap(self._lyr_json)
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<{self.__class__.__name__} @ {self._url}>"
    #----------------------------------------------------------------------
    def __repr__(self):
        return self.__str__()
    #----------------------------------------------------------------------
    @property
    def title(self) -> str:
        """
        The title of the layer used to identify it in places such as the Legend and LayerList widgets.

        :returns: String
        """
        return self._title
    #----------------------------------------------------------------------
    @title.setter
    def title(self, value:str):
        """
        The title of the layer used to identify it in places such as the Legend and LayerList widgets.

        :returns: String
        """
        if self._title != value:
            self._title = value
    #----------------------------------------------------------------------
    @property
    def opacity(self) -> float:
        """
        This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.

        :returns: Float
        """
        return self._opacity
    #----------------------------------------------------------------------
    @opacity.setter
    def opacity(self, value:float):
        """
        This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.

        :returns: Float
        """
        if isinstance(value, (float, int)):
            self._opacity = value
    #----------------------------------------------------------------------
    @property
    def scale(self):
        """Gets/Sets the Min/Max Scale for the layer"""
        return self._min_scale, self._max_scale
    #----------------------------------------------------------------------
    @scale.setter
    def scale(self, scale:tuple):
        """Gets/Sets the Min/Max Scale for the layer"""
        if isinstance(scale, (tuple, list)) and len(scale) == 2:
            self._min_scale, self._max_scale = scale
    #----------------------------------------------------------------------
    @property
    def copyright(self):
        """Copyright information for the layer."""
        return self._copyright
    #----------------------------------------------------------------------
    @copyright.setter
    def copyright(self):
        """Copyright information for the layer."""
        return self._copyright
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self) -> dict:
        """Represents the MapView's JSON format"""
        return {
            "id" : uuid.uuid4().hex,
            "title" : self._title or "Layer",
            "url" : self._url,
            "type" : self._type,
            "minScale" : self.scale[0],
            "maxScale" : self.scale[1],
            "opacity" : self.opacity
        }
    @property
    def _operational_layer_json(self) -> dict:
        """Represents the WebMap's JSON format"""
        return self._lyr_json

###########################################################################
class BaseOpenData(BaseOGC):
    """

    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required string. The web address of the endpoint.
    ---------------     --------------------------------------------------------------------
    gis                 Optional GIS. The `GIS` connection object
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    id                  Optional String. The unique ID of the layer.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ---------------     --------------------------------------------------------------------
    sql_expression      Optional String. Optional query string to apply to the layer when displayed on the widget or web map.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and Layer List widgets.
    ===============     ====================================================================

    """
    _sql = None
    def __init__(self, url, gis=None, **kwargs):
        super(BaseOpenData, self)
        self._url = url
        self._gis = gis
        self._min_scale, self._max_scale = kwargs.pop('scale', (0,0))
        self._title = kwargs.pop("title", "Layer")
        self._opacity = kwargs.pop('opacity', 0)
        self._id = kwargs.pop('id', uuid.uuid4().hex)
        self._copyright = kwargs.pop("copyright", "")
        self._sql = kwargs.pop("sql_expression", None)
    #----------------------------------------------------------------------
    @property
    def sql_expression(self):
        """
        The SQL where clause used to filter features on the client. Only
        the features that satisfy the definition expression are displayed
        in the widget. Setting a definition expression is useful when the
        dataset is large and you don't want to bring all features to the
        client for analysis. The `sql_expressions` may be set when a
        layer is constructed prior to it loading in the view or after it
        has been loaded into the class.

        :return: String
        """
        return self._sql
    #----------------------------------------------------------------------
    @sql_expression.setter
    def sql_expression(self, value):
        """
        The SQL where clause used to filter features on the client. Only
        the features that satisfy the definition expression are displayed
        in the widget. Setting a definition expression is useful when the
        dataset is large and you don't want to bring all features to the
        client for analysis. The `sql_expressions` may be set when a
        layer is constructed prior to it loading in the view or after it
        has been loaded into the class.

        :return: String
        """
        if self._sql != value:
            self._sql = value


import concurrent.futures
import uuid
import datetime 
import logging

_log = logging.getLogger()

def _execute(fn, **kwargs):
    """runs a generic function with optional inputs"""
    if len(kwargs) > 0:
        fn(**kwargs)
    else:
        fn()

            

class Job(object):
    """represents an asynchronous job"""
    _future = None
    _gis = None
    _task_name = None
    _start_time = None
    _end_time = None
    _verbose = None
    #----------------------------------------------------------------------
    def __init__(self, future, task_name, jobid=None, task_url=None, notify=False, gis=None):
        self._start_time = datetime.datetime.now()
        self._task_name = task_name
        self._future = future
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)        
        
        self._end_time = None
        self._url = task_url
        if jobid is None:
            self._jobid = uuid.uuid4().hex
        else:
            self._jobid = jobid
    #----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self._task_name} job {self._jobid}>"
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<{self._task_name} job {self._jobid}>"
    #----------------------------------------------------------------------
    def cancelled(self):
        """
        Return True if the call was successfully cancelled.

        :returns: boolean
        """
        return self._future.cancelled()
    #----------------------------------------------------------------------
    def running(self):
        """
        Return True if the call is currently being executed and cannot be cancelled.

        :returns: boolean
        """
        return self._future.running()
    #----------------------------------------------------------------------
    def done(self):
        """
        Return True if the call was successfully cancelled or finished running.

        :returns: boolean
        """
        return self._future.done()
    #----------------------------------------------------------------------
    @property
    def ellapse_time(self):
        """
        Returns the Ellapse Time for the Job
        """
        if self._end_time:
            return self._end_time - self._start_time
        else:
            return datetime.datetime.now() - self._start_time
    #----------------------------------------------------------------------
    def _set_end_time(self, future):
        """sets the finish time"""
        self._end_time = datetime.datetime.now()
    #----------------------------------------------------------------------
    def _notify(self, future):
        """prints finished method"""
        jobid = self._jobid
        try:
            res = future.result()
            infomsg = '{jobid} finished successfully.'.format(jobid=jobid)
            _log.info(infomsg)
            print(infomsg)
        except Exception as e:
            msg = str(e)
            msg = '{jobid} failed: {msg}'.format(jobid=jobid, msg=msg)
            _log.info(msg)
            print(msg)
    #----------------------------------------------------------------------
    def result(self):
        """returns the job result"""
        return self._future.result()

class GeometryJob(Job):
    _future = None
    _gis = None
    _task_name = None
    _start_time = None
    _end_time = None
    _verbose = None    
    _wkid = None
    def __init__(self, 
                 future:concurrent.futures.Future, 
                 task_name:str, 
                 jobid:str=None, 
                 task_url:str=None, 
                 notify:bool=False, 
                 gis=None,
                 out_wkid:int=None) -> None:
        super(GeometryJob, self)
        self._start_time = datetime.datetime.now()
        self._task_name = task_name
        self._future = future
        self._wkid = out_wkid
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)        
        
        self._end_time = None
        self._url = task_url
        if jobid is None:
            self._jobid = uuid.uuid4().hex
        else:
            self._jobid = jobid
    #----------------------------------------------------------------------
    def result(self):
        """returns the job result"""
        from arcgis.geometry import Geometry
        if self._wkid: # set the output sr to that wkid integer
            sr = {'spatialReference' : { 'wkid' : self._wkid}}
            res = self._future.result()
            if isinstance(res, (list, tuple)):
            
                [g.update(sr) for g in res if not 'spatialReference' in g]
                return [Geometry(g) for g in res]
            elif isinstance(res, dict) and not 'spatialReference' in res:
                res.update(sr)
                return Geometry(res)
            else:
                return res
        else:
            return self._future.result()
    
class ItemStatusJob(Job):
    """represents an Item"""   
    _future = None
    _gis = None
    _task_name = None
    _start_time = None
    _end_time = None
    _verbose = None
    #----------------------------------------------------------------------
    def __init__(self, item, task_name, jobid=None, job_type=None, notify=False, gis=None):
        
        executor =  concurrent.futures.ThreadPoolExecutor(1)
        future = executor.submit(self._status, *(item, jobid, job_type))
        executor.shutdown(False)        
        self._start_time = datetime.datetime.now()
        self._task_name = task_name
        self._future = future
        if notify:
            self._future.add_done_callback(self._notify)
        self._future.add_done_callback(self._set_end_time)        
        
        self._end_time = None
        self._url = None
        if jobid is None:
            self._jobid = uuid.uuid4().hex
        else:
            self._jobid = jobid
    #----------------------------------------------------------------------
    def __repr__(self):
        return f"<{self._task_name} job>"
    #----------------------------------------------------------------------
    def __str__(self):
        return f"<{self._task_name} job>"
    #----------------------------------------------------------------------
    def _status(self, item, job_id, job_type):
        """Processes the asynchronous job"""
        import time
        from arcgis.gis import Item
        
        while item.status(job_id=job_id, job_type=job_type).get('status') in ['processing']:
            status = item.status(job_id=job_id, job_type=job_type).get('status')
            if status in ['failed', 'completed']:
                return item.status(job_id=job_id, job_type=job_type)
            else:
                return item
        return item
    #----------------------------------------------------------------------



#                                  Apache License
#                            Version 2.0, January 2004
#                         http://www.apache.org/licenses/

#    TERMS AND CONDITIONS FOR USE, REPRODUCTION, AND DISTRIBUTION

#    1. Definitions.

#       "License" shall mean the terms and conditions for use, reproduction,
#       and distribution as defined by Sections 1 through 9 of this document.

#       "Licensor" shall mean the copyright owner or entity authorized by
#       the copyright owner that is granting the License.

#       "Legal Entity" shall mean the union of the acting entity and all
#       other entities that control, are controlled by, or are under common
#       control with that entity. For the purposes of this definition,
#       "control" means (i) the power, direct or indirect, to cause the
#       direction or management of such entity, whether by contract or
#       otherwise, or (ii) ownership of fifty percent (50%) or more of the
#       outstanding shares, or (iii) beneficial ownership of such entity.

#       "You" (or "Your") shall mean an individual or Legal Entity
#       exercising permissions granted by this License.

#       "Source" form shall mean the preferred form for making modifications,
#       including but not limited to software source code, documentation
#       source, and configuration files.

#       "Object" form shall mean any form resulting from mechanical
#       transformation or translation of a Source form, including but
#       not limited to compiled object code, generated documentation,
#       and conversions to other media types.

#       "Work" shall mean the work of authorship, whether in Source or
#       Object form, made available under the License, as indicated by a
#       copyright notice that is included in or attached to the work
#       (an example is provided in the Appendix below).

#       "Derivative Works" shall mean any work, whether in Source or Object
#       form, that is based on (or derived from) the Work and for which the
#       editorial revisions, annotations, elaborations, or other modifications
#       represent, as a whole, an original work of authorship. For the purposes
#       of this License, Derivative Works shall not include works that remain
#       separable from, or merely link (or bind by name) to the interfaces of,
#       the Work and Derivative Works thereof.

#       "Contribution" shall mean any work of authorship, including
#       the original version of the Work and any modifications or additions
#       to that Work or Derivative Works thereof, that is intentionally
#       submitted to Licensor for inclusion in the Work by the copyright owner
#       or by an individual or Legal Entity authorized to submit on behalf of
#       the copyright owner. For the purposes of this definition, "submitted"
#       means any form of electronic, verbal, or written communication sent
#       to the Licensor or its representatives, including but not limited to
#       communication on electronic mailing lists, source code control systems,
#       and issue tracking systems that are managed by, or on behalf of, the
#       Licensor for the purpose of discussing and improving the Work, but
#       excluding communication that is conspicuously marked or otherwise
#       designated in writing by the copyright owner as "Not a Contribution."

#       "Contributor" shall mean Licensor and any individual or Legal Entity
#       on behalf of whom a Contribution has been received by Licensor and
#       subsequently incorporated within the Work.

#    2. Grant of Copyright License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       copyright license to reproduce, prepare Derivative Works of,
#       publicly display, publicly perform, sublicense, and distribute the
#       Work and such Derivative Works in Source or Object form.

#    3. Grant of Patent License. Subject to the terms and conditions of
#       this License, each Contributor hereby grants to You a perpetual,
#       worldwide, non-exclusive, no-charge, royalty-free, irrevocable
#       (except as stated in this section) patent license to make, have made,
#       use, offer to sell, sell, import, and otherwise transfer the Work,
#       where such license applies only to those patent claims licensable
#       by such Contributor that are necessarily infringed by their
#       Contribution(s) alone or by combination of their Contribution(s)
#       with the Work to which such Contribution(s) was submitted. If You
#       institute patent litigation against any entity (including a
#       cross-claim or counterclaim in a lawsuit) alleging that the Work
#       or a Contribution incorporated within the Work constitutes direct
#       or contributory patent infringement, then any patent licenses
#       granted to You under this License for that Work shall terminate
#       as of the date such litigation is filed.

#    4. Redistribution. You may reproduce and distribute copies of the
#       Work or Derivative Works thereof in any medium, with or without
#       modifications, and in Source or Object form, provided that You
#       meet the following conditions:

#       (a) You must give any other recipients of the Work or
#           Derivative Works a copy of this License; and

#       (b) You must cause any modified files to carry prominent notices
#           stating that You changed the files; and

#       (c) You must retain, in the Source form of any Derivative Works
#           that You distribute, all copyright, patent, trademark, and
#           attribution notices from the Source form of the Work,
#           excluding those notices that do not pertain to any part of
#           the Derivative Works; and

#       (d) If the Work includes a "NOTICE" text file as part of its
#           distribution, then any Derivative Works that You distribute must
#           include a readable copy of the attribution notices contained
#           within such NOTICE file, excluding those notices that do not
#           pertain to any part of the Derivative Works, in at least one
#           of the following places: within a NOTICE text file distributed
#           as part of the Derivative Works; within the Source form or
#           documentation, if provided along with the Derivative Works; or,
#           within a display generated by the Derivative Works, if and
#           wherever such third-party notices normally appear. The contents
#           of the NOTICE file are for informational purposes only and
#           do not modify the License. You may add Your own attribution
#           notices within Derivative Works that You distribute, alongside
#           or as an addendum to the NOTICE text from the Work, provided
#           that such additional attribution notices cannot be construed
#           as modifying the License.

#       You may add Your own copyright statement to Your modifications and
#       may provide additional or different license terms and conditions
#       for use, reproduction, or distribution of Your modifications, or
#       for any such Derivative Works as a whole, provided Your use,
#       reproduction, and distribution of the Work otherwise complies with
#       the conditions stated in this License.

#    5. Submission of Contributions. Unless You explicitly state otherwise,
#       any Contribution intentionally submitted for inclusion in the Work
#       by You to the Licensor shall be under the terms and conditions of
#       this License, without any additional terms or conditions.
#       Notwithstanding the above, nothing herein shall supersede or modify
#       the terms of any separate license agreement you may have executed
#       with Licensor regarding such Contributions.

#    6. Trademarks. This License does not grant permission to use the trade
#       names, trademarks, service marks, or product names of the Licensor,
#       except as required for reasonable and customary use in describing the
#       origin of the Work and reproducing the content of the NOTICE file.

#    7. Disclaimer of Warranty. Unless required by applicable law or
#       agreed to in writing, Licensor provides the Work (and each
#       Contributor provides its Contributions) on an "AS IS" BASIS,
#       WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
#       implied, including, without limitation, any warranties or conditions
#       of TITLE, NON-INFRINGEMENT, MERCHANTABILITY, or FITNESS FOR A
#       PARTICULAR PURPOSE. You are solely responsible for determining the
#       appropriateness of using or redistributing the Work and assume any
#       risks associated with Your exercise of permissions under this License.

#    8. Limitation of Liability. In no event and under no legal theory,
#       whether in tort (including negligence), contract, or otherwise,
#       unless required by applicable law (such as deliberate and grossly
#       negligent acts) or agreed to in writing, shall any Contributor be
#       liable to You for damages, including any direct, indirect, special,
#       incidental, or consequential damages of any character arising as a
#       result of this License or out of the use or inability to use the
#       Work (including but not limited to damages for loss of goodwill,
#       work stoppage, computer failure or malfunction, or any and all
#       other commercial damages or losses), even if such Contributor
#       has been advised of the possibility of such damages.

#    9. Accepting Warranty or Additional Liability. While redistributing
#       the Work or Derivative Works thereof, You may choose to offer,
#       and charge a fee for, acceptance of support, warranty, indemnity,
#       or other liability obligations and/or rights consistent with this
#       License. However, in accepting such obligations, You may act only
#       on Your own behalf and on Your sole responsibility, not on behalf
#       of any other Contributor, and only if You agree to indemnify,
#       defend, and hold each Contributor harmless for any liability
#       incurred by, or claims asserted against, such Contributor by reason
#       of your accepting any such warranty or additional liability.

#    END OF TERMS AND CONDITIONS

#    APPENDIX: How to apply the Apache License to your work.

#       To apply the Apache License to your work, attach the following
#       boilerplate notice, with the fields enclosed by brackets "[]"
#       replaced with your own identifying information. (Don't include
#       the brackets!)  The text should be enclosed in the appropriate
#       comment syntax for the file format. We also recommend that a
#       file or class name and description of purpose be included on the
#       same "printed page" as the copyright notice for easier
#       identification within third-party archives.

#    Copyright [yyyy] [name of copyright owner]

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.


from typing import List, Tuple
from torch import nn
import torch.nn.functional as F
from functools import partial
from fastai.basic_train import LearnerCallback
import numpy as np
import torch
from arcgis.learn._utils.classified_tiles import calculate_intersection, calculate_union, confusion_matrix, calculate_precision, calculate_recall, calculate_f1, calculate_metrics



def expand_outputs(preds,trues):
    if preds.shape == trues.shape:
        return trues
    encoded_trues=preds.detach()*0
    encoded_trues.scatter_(1, trues, 1)
    return encoded_trues


def dice(preds, *targs, eps=1e-8, iou=False, soft=False, ignore_classes=[0], weighted=False, **kwargs, ):
    """
    Calculates dice coefficient over a batch.
    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    preds                   Required torch.tensor.
                            Predictions form a segmentation model.
                            file.
    ---------------------   -------------------------------------------
    targs                   Required torch.tensor.
                            A batch of ground truth segmentation mask.
    ---------------------   -------------------------------------------
    eps                     Optional float.
                            A very small floating point number to avoid deviding
                            by zero errors.
                            Default : 1e-08
    ---------------------   -------------------------------------------
    iou                     Optional Bool.
                            False: Compute Dice coefficient.
                            True:  Compute mean IOU
    ---------------------   -------------------------------------------
    ignore_classes=[0],     Optional List.
                            A list of classes(ignore_mapped_class) on which needs to be skipped for
                            dice calculation.
    ---------------------   -------------------------------------------
    weighted=False          Optional Bool.
                            Not implemented
                            Default False.
    =====================   ===========================================
    :returns: Dice Coefficient->Rank 0 torch.tensor
    """
    if (
            isinstance(preds, (List, Tuple))
            and isinstance(targs, (List, Tuple))
            and len(preds) == len(targs)
    ):
        return torch.FloatTensor(
            [
                dice(pred, targ, ignore_classes=ignore_class)
                for pred, targ, ignore_class in zip(preds, targs, ignore_classes)
            ]
        )
    if isinstance(preds, (List, Tuple)):
        preds = preds[0]
    if isinstance(targs, (List, Tuple)) and len(targs) == 1:
        targs = targs[0].long()
    if len(targs.size()) == 3:
        targs = targs.unsqueeze(1)
    num = preds.size(0)
    classes = preds.size(1)
    preds = F.softmax(preds, dim=1)
    keep_classes = [v for v in range(classes) if v not in ignore_classes]
    encoded_targs = expand_outputs(preds, targs)

    if weighted:
        # need to work on a weighted metric and loss function
        pass

    if not soft:  # for metric calculation
        preds = preds.argmax(dim=1)
        preds = torch.nn.functional.one_hot(preds, classes).permute(0, 3, 1, 2)

    preds = preds.contiguous().float().view(num, classes, -1)[:, keep_classes, :].view(num, -1)
    encoded_targs = encoded_targs.float().view(num, classes, -1)[:, keep_classes, :].view(num, -1)
    intersection = calculate_intersection(preds, encoded_targs, mode='mean')
    union = calculate_union(preds, encoded_targs, intersection, mode='mean')

    if not iou:
        score = 2. * intersection / (union + intersection + eps)
    else:
        score = intersection / (union + eps)

    score[union == 0.] = 1.
    mean_per_img = score.mean(dim=0)
    return mean_per_img.mean()

# Below loss functions are implementations from https://github.com/simongrest/farm-pin-crop-detection-challenge 
class FocalLoss(nn.Module):
    def __init__(self, crit, alpha=1, gamma=2):
        super().__init__()
        self.alpha = alpha
        self.gamma = gamma
        self.crit = crit
    
    def forward(self, inputs, targets, reduction='mean'):
        loss = self.crit(inputs, targets)
        pt = torch.exp(-loss)
        F_loss = self.alpha * (1-pt)**self.gamma * loss

        if reduction is None:
            return F_loss
        else:
            return torch.mean(F_loss)

class DiceLoss(nn.Module):
    def __init__(self, crit, pct, weighted_dice):
        super().__init__()
        self.crit = crit
        self.pct = pct
        self.weighted_dice = weighted_dice
    
    def forward(self, inputs, targets, **kwargs):
        weighted = self.weighted_dice
        if weighted:
            ignore_mapped_class = kwargs.get('ignore_mapped_class', [])
        else:
            ignore_mapped_class = kwargs.get('ignore_mapped_class', [0])

        dice_complement = self.crit(inputs, targets)
        if isinstance(inputs, tuple): # handling for aux_loss=True 
            inputs=inputs[0]
        dice_c=dice(inputs, targets, soft=True, weighted=weighted, ignore_classes=ignore_mapped_class)
        
        dice_loss = torch.clamp(1 - dice_c,0,1)
        loss = (1-self.pct) * dice_complement + self.pct * dice_loss
        return loss

class MixUpCallback(LearnerCallback):
    "Callback that creates the mixed-up input and target."
    def __init__(self, learn, alpha=0.4, stack_x=False, stack_y=True):
        super().__init__(learn)
        self.alpha,self.stack_x,self.stack_y = alpha,stack_x,stack_y
    
    def on_train_begin(self, **kwargs):
        if self.stack_y: self.learn.loss_func = MixUpLoss(self.learn.loss_func)
        
    def on_batch_begin(self, last_input, last_target, train, **kwargs):
        "Applies mixup to `last_input` and `last_target` if `train`."
        if not train: return
        lambd = np.random.beta(self.alpha, self.alpha, last_target.size(0))
        lambd = np.concatenate([lambd[:,None], 1-lambd[:,None]], 1).max(1)
        lambd = last_input.new(lambd)

        shuffle = torch.randperm(last_target.size(0)).to(last_input.device)
        x1, y1 = last_input[shuffle], last_target[shuffle]
        if self.stack_x:
            new_input = [last_input, last_input[shuffle], lambd]
        else: 
            out_shape = [lambd.size(0)] + [1 for _ in range(len(x1.shape) - 1)]
            new_input = (last_input * lambd.view(out_shape) + x1 * (1-lambd).view(out_shape))
        if self.stack_y:
            
            new_lambd = torch.distributions.utils.broadcast_all(lambd[:,None,None,None], last_target)[0]
            
            #new_target = torch.cat([last_target[:,None].float(), y1[:,None].float(), new_lambd[:,None].float()], 1)
            new_target = torch.stack([last_target.float(), y1.float(), new_lambd.float()], 1)
        else:
            if len(last_target.shape) == 2:
                lambd = lambd.unsqueeze(1).float()
            new_target = last_target.float() * lambd + y1.float() * (1-lambd)
            
        return {'last_input': new_input, 'last_target': new_target}  
    
    def on_train_end(self, **kwargs):
        if self.stack_y: self.learn.loss_func = self.learn.loss_func.get_old()
        

class MixUpLoss(nn.Module):
    "Adapt the loss function `crit` to go with mixup."
    
    def __init__(self, crit, reduction='mean'):
        super().__init__()
        if hasattr(crit, 'reduction'): 
            self.crit = crit
            self.old_red = crit.reduction
            setattr(self.crit, 'reduction', 'none')
        else: 
            self.crit = partial(crit, reduction='none')
            self.old_crit = crit
        self.reduction = reduction
        
    def forward(self, output, target):
        if len(target.size()) >= 5:
            loss1, loss2 = self.crit(output,target[:,0].long()), self.crit(output,target[:,1].long())
            lambd = target[:,2].contiguous().view(-1)
            d = (loss1 * lambd  + loss2 * (1-lambd)).mean()
        else:  d = self.crit(output, target)
        if self.reduction == 'mean': return d.mean()
        elif self.reduction == 'sum':            return d.sum()
        return d
    
    def get_old(self):
        if hasattr(self, 'old_crit'):  return self.old_crit
        elif hasattr(self, 'old_red'): 
            setattr(self.crit, 'reduction', self.old_red)
            return self.crit



from functools import partial
from pathlib import Path
import sys
import json
import warnings
import traceback
from ..models._arcgis_model import ArcGISModel, model_characteristics_folder

HAS_NUMPY = True
HAS_FASTAI = True

try:
    import torch
    import torch.nn as nn
    import pandas as pd
    from fastai.text.transform import Tokenizer
    from fastprogress.fastprogress import progress_bar
    from fastai.basic_train import Learner, DatasetType
    from fastai.train import to_fp16
    from fastai.metrics import accuracy, error_rate, accuracy_thresh
    # from transformers import AdamW
    from transformers import AutoTokenizer, AutoConfig
    from .._utils.env  import LAMBDA_TEXT_CLASSIFICATION
    if not LAMBDA_TEXT_CLASSIFICATION:
        from sklearn.metrics import classification_report
    from ._arcgis_transformer import ModelBackbone, infer_model_type
    from .._utils.common import _get_emd_path
    from .._utils.text_data import TextDataObject, save_data_in_model_metrics_html, copy_metrics
    from .._utils.text_transforms import TransformersBaseTokenizer, TransformersVocab
    from ._transformer_text_classifier import TransformerForTextClassification, backbone_models_reverse_map, \
        transformer_architectures, transformer_seq_length
    from transformers import logging
except Exception as e:
    import_exception = "\n".join(traceback.format_exception(type(e), e, e.__traceback__))
    HAS_FASTAI = False
    from ._transformer_text_classifier import transformer_architectures, transformer_seq_length

    class TransformerForTextClassification:
        _supported_backbones = transformer_architectures
else:
    warnings.filterwarnings("ignore", category=UserWarning, module='fastai')


try:
    import numpy as np
    warnings.filterwarnings("ignore", category=np.VisibleDeprecationWarning) 
except:
    HAS_NUMPY = False


class TextClassifier(ArcGISModel):
    """
    Creates a TextClassifier Object.
    Based on the Hugging Face transformers library

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required text data object, returned from
                            prepare_textdata function.
    ---------------------   -------------------------------------------
    backbone                Optional string. Specifying the HuggingFace
                            transformer model name to be used to train the
                            classifier. Default set to `bert-base-cased`.

                            To learn more about the available models or
                            choose models that are suitable for your dataset,
                            kindly visit:- https://huggingface.co/transformers/pretrained_models.html
    =====================   ===========================================

    **kwargs**

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    verbose                 Optional string. Default set to `error`. The
                            log level you want to set. It means the amount
                            of information you want to display while training
                            or calling the various methods of this class.
                            Allowed values are - `debug`, `info`, `warning`,
                            `error` and `critical`.
    ---------------------   -------------------------------------------
    seq_len                 Optional Integer. Default set to 512. Maximum
                            sequence length (at sub-word level after tokenization)
                            of the training data to be considered for training
                            the model.
    ---------------------   -------------------------------------------
    thresh                  Optional Float. This parameter is used to set
                            the threshold value to pick labels in case of
                            multi-label text classification problem. Default
                            value is set to 0.25
    ---------------------   -------------------------------------------
    mixed_precision         Optional Bool. Default set to False. If set
                            True, then mixed precision training is used
                            to train the model
    ---------------------   -------------------------------------------
    pretrained_path         Optional String. Path where pre-trained model
                            is saved. Accepts a Deep Learning Package
                            (DLPK) or Esri Model Definition(EMD) file.
    =====================   ===========================================

    :returns: `TextClassifier` Object
    """

    # supported transformer backbones
    supported_backbones = transformer_architectures

    def __init__(self, data, backbone="bert-base-cased", **kwargs):
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error
            _raise_fastai_import_error(import_exception=import_exception)

        self.logger = logging.get_logger()
        if kwargs.get('verbose', None):
            self.logger.setLevel(kwargs.get('verbose').upper())
        else:
            self.logger.setLevel(logging.ERROR)

        model_backbone = ModelBackbone(backbone)
        super().__init__(data, model_backbone)
        self.is_multilabel_problem = False
        self.thresh = kwargs.get('thresh', 0.25)
        self._mixed_precision = kwargs.get('mixed_precision', False)
        self._seq_len = kwargs.get('seq_len', transformer_seq_length)
        self._create_text_learner_object(
            data, backbone, kwargs.get('pretrained_path', None), mixed_precision=self._mixed_precision, seq_len=self._seq_len)

        self.learn.model = self.learn.model.to(self._device)
        layer_groups = self.learn.model.get_layer_groups()
        self.learn.split(layer_groups)
        self._freeze()

    def _create_text_learner_object(self, data, backbone, pretrained_path=None, mixed_precision=False,
                                    seq_len=transformer_seq_length):
        model_type = infer_model_type(backbone, transformer_architectures)
        self.logger.info(f"Inferred Backbone: {model_type}")
        pretrained_model_name = backbone

        transformer_tokenizer = AutoTokenizer.from_pretrained(pretrained_model_name)

        # pad_first = bool(model_type in ['xlnet'])
        pad_first = True if transformer_tokenizer.padding_side == 'left' else False
        pad_idx = transformer_tokenizer.pad_token_id

        base_tokenizer = TransformersBaseTokenizer(pretrained_tokenizer=transformer_tokenizer, seq_len=seq_len)
        tokenizer = Tokenizer(tok_func=base_tokenizer, pre_rules=[], post_rules=[])
        if sys.platform == 'win32': tokenizer.n_cpus = 1
        vocab = TransformersVocab(tokenizer=transformer_tokenizer)

        if data._is_empty or data._backbone != backbone:
            self.logger.info('Creating DataBunch')
            data._prepare_databunch(tokenizer=tokenizer, vocab=vocab, pad_first=pad_first,
                                    pad_idx=pad_idx, backbone=backbone, logger=self.logger)

        databunch = data.get_databunch()

        config = AutoConfig.from_pretrained(pretrained_model_name)
        config.label2id = databunch.train_ds.c2i
        config.id2label = {y: x for x, y in config.label2id.items()}

        if pretrained_path is not None: pretrained_path = str(_get_emd_path(pretrained_path))

        model = TransformerForTextClassification(
            architecture=model_type,
            pretrained_model_name=pretrained_model_name,
            config=config,
            pretrained_model_path=pretrained_path,
            seq_len=seq_len
        )
        model.init_model()
        # opt_func = partial(AdamW, correct_bias=False)

        self.is_multilabel_problem = True if len(data._label_cols) > 1 else False
        if self.is_multilabel_problem:
            accuracy_multi = partial(accuracy_thresh, thresh=self.thresh)
            accuracy_multi.__name__ = "accuracy"
            metrics = [accuracy_multi]
            loss_func = nn.BCEWithLogitsLoss()
            # self.learn = Learner(databunch, model, opt_func=opt_func, loss_func=loss_func, metrics=metrics)
            self.learn = Learner(databunch, model, loss_func=loss_func, metrics=metrics)
        else:
            metrics = [accuracy, error_rate]

            # self.learn = Learner(databunch, model, opt_func=opt_func, metrics=metrics)
            self.learn = Learner(databunch, model, metrics=metrics)

        if pretrained_path is not None:
            self.load(pretrained_path)

        if mixed_precision:
            if model_type in ["xlnet", "mobilebert"]:
                error_message = (
                    f"Mixed precision training is not supported for transformer model - {model_type.upper()}."
                    "\nKindly turn off the `mixed_precision` flag to use this model in its default mode,"
                    f" or choose a different transformer architectures from - {transformer_architectures}")
                raise Exception(error_message)
            self.logger.info("Converting model to 16 Bit Floating Point precision")
            self.learn = to_fp16(self.learn)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @classmethod
    def available_backbone_models(cls, architecture):
        """
        Get available models for the given transformer backbone

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        architecture            Required string. name of the transformer
                                backbone one wish to use. To learn more about
                                the available models or choose models that are
                                suitable for your dataset, kindly visit:-
                                https://huggingface.co/transformers/pretrained_models.html
        =====================   ===========================================

        :returns: a tuple containing the available models for the given transformer backbone
        """
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error
            _raise_fastai_import_error(import_exception=import_exception)
        return TransformerForTextClassification._available_backbone_models(architecture)

    def _freeze(self):
        """
        Freeze up to last layer group to train only the last layer group of the model.
        """
        self.learn.freeze()

    @classmethod
    def from_model(cls, emd_path, data=None):
        """
        Creates an TextClassifier model object from a Deep Learning
        Package(DLPK) or Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from `prepare_textdata` function or None for
                                inferencing.
        =====================   ===========================================

        :returns: `TextClassifier` model Object
        """
        if not HAS_FASTAI:
            from .._data import _raise_fastai_import_error
            _raise_fastai_import_error(import_exception=import_exception)

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)

        pretrained_model = emd["PretrainedModel"]
        mixed_precision = emd["MixedPrecisionTraining"]
        text_cols = emd["TextColumns"]
        label_cols = emd["LabelColumns"]
        class_labels = list(emd["Label2Id"].keys())
        is_multilabel_problem = emd["IsMultilabelClassificationProblem"]
        thresh = emd.get("Threshold")
        seq_len = emd.get("SequenceLength", transformer_seq_length)

        data_is_none = False
        if data is None:
            data_is_none = True
            data = TextDataObject(task="classification")
            data._backbone = pretrained_model
            data.create_empty_object_for_classification(text_cols, label_cols, class_labels, is_multilabel_problem)
            data.emd, data.emd_path = emd, emd_path.parent
        cls_object = cls(data, pretrained_model, pretrained_path=str(emd_path),
                         mixed_precision=mixed_precision, thresh=thresh, seq_len=seq_len)
        if data_is_none: cls_object._data._is_empty = True
        return cls_object

    def load(self, name_or_path):
        """
        Loads a saved TextClassifier model from disk.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        =====================   ===========================================
        """
        if '\\' in str(name_or_path) or '/' in str(name_or_path):
            name_or_path = str(_get_emd_path(name_or_path))
        else:
            name_or_path = Path('models') / name_or_path
            name_or_path = str(_get_emd_path(name_or_path))
        return super().load(name_or_path)

    def save(self, name_or_path, framework='PyTorch', publish=False, gis=None, compute_metrics=True,
             save_optimizer=False, **kwargs):
        """
        Saves the model weights, creates an Esri Model Definition and Deep
        Learning Package zip for deployment.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        name_or_path            Required string. Folder path to save the model.
        ---------------------   -------------------------------------------
        framework               Optional string. Defines the framework of the
                                model. (Only supported by ``SingleShotDetector``, currently.)
                                If framework used is ``TF-ONNX``, ``batch_size`` can be
                                passed as an optional keyword argument.

                                Framework choice: 'PyTorch' and 'TF-ONNX'
        ---------------------   -------------------------------------------
        publish                 Optional boolean. Publishes the DLPK as an item.
        ---------------------   -------------------------------------------
        gis                     Optional GIS Object. Used for publishing the item.
                                If not specified then active gis user is taken.
        ---------------------   -------------------------------------------
        compute_metrics         Optional boolean. Used for computing model
                                metrics.
        ---------------------   -------------------------------------------
        save_optimizer          Optional boolean. Used for saving the model-optimizer
                                state along with the model. Default is set to False.
        ---------------------   -------------------------------------------
        kwargs                  Optional Parameters:
                                Boolean `overwrite` if True, it will overwrite
                                the item on ArcGIS Online/Enterprise, default False.
                                Boolean `zip_files` if True, it will create the Deep
                                Learning Package (DLPK) file while saving the model.
        =====================   ===========================================

        :returns: the qualified path at which the model is saved
        """
        from ..models._arcgis_model import _create_zip
        zip_files = kwargs.pop('zip_files', True)
        overwrite = kwargs.pop('overwrite', False)
        path = super().save(name_or_path, framework, publish=False, gis=None, compute_metrics=compute_metrics,
                            save_optimizer=save_optimizer, zip_files=False, **kwargs)

        self._save_df_to_html(path)

        if zip_files:
            _create_zip(path.name, str(path))

        if publish:
            self._publish_dlpk((path/path.stem).with_suffix('.dlpk'), gis=gis, overwrite=overwrite)

        return Path(path)

    @property
    def _model_metrics(self):
        from IPython.utils import io
        with io.capture_output() as captured:
            metrics = {"Accuracy": self.accuracy()}
            per_class_metric_df = self.metrics_per_label()
            metrics["MetricsPerLabel"] = json.dumps(per_class_metric_df.transpose().to_dict())
        return metrics

    def _get_emd_params(self, save_inference_file):
        _emd_template = {}
        is_multilabel_problem = True if len(self._data._label_cols) > 1 else False
        _emd_template["Architecture"]= self.learn.model._transformer_architecture
        _emd_template["PretrainedModel"]= self.learn.model._transformer_pretrained_model_name
        _emd_template["ModelType"] = "Transformer"
        _emd_template["MixedPrecisionTraining"] = self._mixed_precision
        _emd_template["TextColumns"] = self._data._text_cols
        _emd_template["LabelColumns"] = self._data._label_cols
        _emd_template["Label2Id"] = self.learn.model._config.label2id
        _emd_template["SequenceLength"] = self._seq_len
        if is_multilabel_problem: _emd_template["Threshold"] = self.thresh
        _emd_template["IsMultilabelClassificationProblem"] = is_multilabel_problem
        return _emd_template

    def show_results(self, rows=5, **kwargs):
        """
        Prints the rows of the dataframe with target and prediction columns.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        rows                    Optional Integer.
                                Number of rows to print.
        =====================   ===========================================

        :returns: dataframe
        """
        self._check_requisites()
        if kwargs.get("thresh") is None and self.is_multilabel_problem:
            kwargs.update({"thresh": self.thresh})
        return self.learn.show_results(rows=rows, **kwargs)

    def accuracy(self):
        """
        Calculates the following  metric:
            * accuracy:   the number of correctly predicted labels in the validation set
                          divided by the total number of items in the validation set
        :returns: a floating point number depicting the accuracy of the classification model.
        """
        try:
            self._check_requisites()
        except Exception as e:
            acc = self._data.emd.get('Accuracy')
            if acc: return acc
            else:   self.logger.error("Metric not found in the loaded model")
        else:
            if not HAS_NUMPY:
                self.logger.error("This function requires numpy.")
                return
            if hasattr(self.learn, 'recorder'):
                metrics_names = self.learn.recorder.metrics_names
                metrics_values = self.learn.recorder.metrics
                if len(metrics_names) > 0 and len(metrics_values) > 0:
                    metrics = {x: round(metrics_values[-1][i].item(), 4) for i, x in enumerate(metrics_names)}
                    metric = metrics["accuracy"]
                else:
                    metric = self._calculate_model_metric()
            else:
                metric = self._calculate_model_metric()
            return metric

    def _calculate_model_metric(self):
        self.logger.info("Calculating Model Metrics")
        validation_dataframe = self._data._valid_df

        if self.is_multilabel_problem:
            predictions = [x[2] for x in self.predict(validation_dataframe[self._data._text_cols].tolist())]
            labels = [[int(getattr(item, column)) for column in self._data._label_cols]
                      for idx, item in validation_dataframe.iterrows()]

            metric = accuracy_thresh \
                (torch.tensor(predictions), torch.tensor(labels), thresh=self.thresh, sigmoid=False).item()
        else:
            predictions = [x[1] for x in self.predict(validation_dataframe[self._data._text_cols].tolist())]
            labels = [x[0] for x in validation_dataframe[self._data._label_cols].values]
            metric = round((np.sum(np.array(predictions) == labels) / len(labels)), 4)

        return metric

    def _predict(self, text, thresh=None):
        if thresh is None: thresh = self.thresh
        result = self.learn.model.predict_class(text, self._device, self.is_multilabel_problem, thresh)
        return result

    def predict(self, text_or_list, show_progress=True, thresh=None):
        """
        Predicts the class label(s) for the input text

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        text_or_list            Required String or List. text or a list of
                                texts for which we wish to find the class label(s).
        ---------------------   -------------------------------------------
        show_progress           optional Bool. If set to True, will display a
                                progress bar depicting the items processed so far.
                                Applicable only when a list of text is passed
        ---------------------   -------------------------------------------
        thresh                  Optional Float. The threshold value set to get
                                the class label(s). Applicable only for multi-label
                                classification task. Default is the value set
                                during the model creation time, otherwise the value
                                of 0.25 is set. 
        =====================   ===========================================

        :returns: * In case of single label classification problem, a tuple containing
                  the text, its predicted class label and the confidence score.

                  * In case of multi label classification problem, a tuple containing
                  the text, its predicted class labels, a list containing 1's for the
                  predicted labels, 0's otherwise and list containing a score for each label
        """
        if self.is_multilabel_problem is False and thresh is not None:
            self.logger.error("Passing a threshold value for non multi-label classification task "
                           "will not have any affect on the predicting the class label")

        if isinstance(text_or_list, (list, tuple, np.ndarray)):
            if show_progress:
                preds = []

                for i in progress_bar(range(len(text_or_list))):
                    prediction = self._predict(text_or_list[i], thresh)
                    preds.append(prediction)
            else:
                preds = [self._predict(x, thresh) for x in text_or_list]

            result = [(text, *pred) for text, pred in zip(text_or_list, preds)]
            return result
        else:
            preds = self._predict(text_or_list, thresh)
            return (text_or_list, *preds)

    def _save_df_to_html(self, path):
        if getattr(self._data, '_is_empty', False):
            copy_metrics(self._data.emd_path, path, model_characteristics_folder)
            return
        validation_dataframe = self._data._valid_df.sample(n=5)
        if self.is_multilabel_problem:
            predictions = [x[1] for x in self.predict(validation_dataframe[self._data._text_cols].tolist(),
                                                          show_progress=False)]
            labels = [";".join([column for column in self._data._label_cols if int(getattr(item, column))])
                      for idx, item in validation_dataframe.iterrows()]

        else:
            predictions = [x[1] for x in self.predict(validation_dataframe[self._data._text_cols].tolist(),
                                                          show_progress=False)]
            labels = [x[0] for x in validation_dataframe[self._data._label_cols].values]



        new_df = pd.DataFrame(validation_dataframe[self._data._text_cols].values, columns=["source"])
        new_df["target"] = labels
        new_df["predictions"] = predictions

        df_str = new_df.to_html(index=False, justify='left').replace(">\n", ">")

        msg = "<p><b>Sample Results</b></p>"

        text = f"\n\n{msg}\n\n{df_str}"

        save_data_in_model_metrics_html(text, path, model_characteristics_folder)

    def metrics_per_label(self):
        """
        :returns: precision, recall and f1 score for each label in the classification model.
        """
        try:
            self._check_requisites()
        except Exception as e:
            metrics_per_label = self._data.emd.get('MetricsPerLabel')
            if metrics_per_label:
                metrics_per_label = json.loads(metrics_per_label)
                return self._create_dataframe_from_dict(metrics_per_label)
            else: self.logger.error("Metric not found in the loaded model")
        else:
            validation_dataframe = self._data._valid_df
            if self.is_multilabel_problem:
                predictions = [x[2] for x in self.predict(validation_dataframe[self._data._text_cols].tolist())]
                labels = [[int(getattr(item, column)) for column in self._data._label_cols] for idx, item in
                          validation_dataframe.iterrows()]
                target_names = self._data._label_cols
                output_dict = classification_report(labels, predictions, target_names=target_names,
                                               zero_division=1, output_dict=True)
            else:
                predictions = [x[1] for x in self.predict(validation_dataframe[self._data._text_cols].tolist())]
                labels = [x[0] for x in validation_dataframe[self._data._label_cols].values]
                target_names = self.learn.model._config.label2id.keys()
                output_dict = classification_report(labels, predictions, target_names=target_names,
                                               zero_division=1, output_dict=True)

            return self._create_dataframe_from_dict(output_dict)

    @staticmethod
    def _create_dataframe_from_dict(out_dict):
        out_dict.pop("accuracy", None)
        out_dict.pop("micro avg", None)
        out_dict.pop("macro avg", None)
        out_dict.pop("samples avg", None)
        out_dict.pop("weighted avg", None)
        df = pd.DataFrame(out_dict)
        # df.drop("support", inplace=True)
        dataframe = df.T.round(4)
        column_mappings = {'precision': 'Precision_score', 'recall': 'Recall_score',
                           'f1-score': 'F1_score', 'support': 'Support'}
        dataframe.rename(columns=column_mappings, inplace=True)
        return dataframe

    def get_misclassified_records(self):
        """
        :returns: get misclassified records for this classification model.
        """
        self._check_requisites()
        validation_dataframe = self._data._valid_df
        misclassified_records, text_col, label_col = [], self._data._text_cols, self._data._label_cols
        if self.is_multilabel_problem:
            predictions = [x[1] for x in self.predict(validation_dataframe[self._data._text_cols].tolist())]
            labels = [";".join([column for column in self._data._label_cols if int(getattr(item, column))])
                      for idx, item in validation_dataframe.iterrows()]
            for index, row in validation_dataframe.iterrows():
                label, prediction = labels[index], predictions[index]
                if label != prediction:
                    misclassified_records.append((row[text_col], label, prediction))
        else:
            predictions = [x[1] for x in self.predict(validation_dataframe[self._data._text_cols].tolist())]
            label_col = label_col if isinstance(label_col, (str, bytes)) else label_col[0]
            for index, row in validation_dataframe.iterrows():
                label, prediction =  row[label_col], predictions[index]
                if label != prediction:
                    misclassified_records.append((row[text_col], label, prediction))

        return pd.DataFrame(misclassified_records, columns=[text_col, "Target", "Prediction"])



import json
import uuid
from urllib.parse import (urlencode, urlparse, urlunparse,
                          parse_qs, ParseResult)
import xml.etree.cElementTree as ET
from io import BytesIO, StringIO

from arcgis.gis import GIS
from arcgis import env as _env
from arcgis._impl.common._mixins import PropertyMap

from ._base import BaseOGC

###########################################################################
class WMTSLayer(BaseOGC):
    """
    Represents a Web Map Tile Service, which is an OGC web service endpoint.


    ===============     ====================================================================
    **Argument**        **Description**
    ---------------     --------------------------------------------------------------------
    url                 Required string. The web address of the endpoint.
    ---------------     --------------------------------------------------------------------
    version             Optional String. The version number of the WMTS service.  The default is `1.0.0`
    ---------------     --------------------------------------------------------------------
    gis                 Optional GIS. The GIS used to reference the service by. The arcgis.env.active_gis is used if not specified.
    ---------------     --------------------------------------------------------------------
    copyright           Optional String. Describes limitations and usage of the data.
    ---------------     --------------------------------------------------------------------
    opacity             Optional Float.  This value can range between 1 and 0, where 0 is 100 percent transparent and 1 is completely opaque.
    ---------------     --------------------------------------------------------------------
    scale               Optional Tuple. The min/max scale of the layer where the positions are: (min, max) as float values.
    ---------------     --------------------------------------------------------------------
    title               Optional String. The title of the layer used to identify it in places such as the Legend and Layer List widgets.
    ===============     ====================================================================



    """
    _gis = None
    _con = None
    _url = None
    _reader = None
    _cap_reader = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url, version='1.0.0', gis=None, **kwargs):
        super(WMTSLayer,self)
        if gis:
            gis = gis
        elif gis is None and _env.active_gis:
            gis = _env.active_gis
        else:
            gis = GIS()
        assert isinstance(gis, GIS)
        self._id = kwargs.pop('id', uuid.uuid4().hex)
        self._version = version
        self._con = gis._con
        self._title = kwargs.pop("title", "WMTS Layer")
        self._gis = gis
        if url[-1] == "/":
            url = url[:-1]
        self._url = url
        self._add_token = str(self._con._auth).lower() == "builtin"
        self._min_scale, self._max_scale = kwargs.pop('scale', (0,0))
        self._opacity = kwargs.pop('opacity', 1)
        self._type = "WebTiledLayer"
    #----------------------------------------------------------------------
    @property
    def properties(self):
        """
        Returns the properties of the Layer.
        
        :returns: PropertyMap
        """        
        if self._properties is None:
            from arcgis._impl.common._mixins import PropertyMap
            if self._add_token:
                url = self._capabilities_url(service_url=self._url, vendor_kwargs={'token' : self._con.token})
            else:
                url = self._capabilities_url(service_url=self._url)
            text = self._con.get(url, {}, try_json=False, add_token=False)
            if text.find("Invalid Token") > -1 or text.find("Get Token") > -1:
                url = self._capabilities_url(service_url=self._url)
                text = self._con.get(url, {}, try_json=False, add_token=False)
            elif text.lower().find("<html>") > -1:
                url = self._capabilities_url(service_url=self._url)
                text = self._con.get(url, {}, try_json=False, add_token=False)
            elif text.lower().find("<?xml version=") > -1:
                pass
            else:
                raise Exception("Could not connect to the WebMap Tile Service")
            sss = BytesIO()
            sss.write(text.encode())
            sss.seek(0)
            tree = ET.XML(text=sss.read())
            d = self._xml_to_dictionary(tree)
            self._properties = PropertyMap(d)
        return self._properties
    #----------------------------------------------------------------------
    def _capabilities_url(self, service_url, vendor_kwargs=None):
        """Return a capabilities url
        """
        pieces = urlparse(service_url)
        args = parse_qs(pieces.query)
        if 'service' not in args:
            args['service'] = 'WMTS'
        if 'request' not in args:
            args['request'] = 'GetCapabilities'
        if 'version' not in args:
            args['version'] = self._version
        if vendor_kwargs:
            args.update(vendor_kwargs)
        query = urlencode(args, doseq=True)
        pieces = ParseResult(pieces.scheme, pieces.netloc,
                             pieces.path, pieces.params,
                             query, pieces.fragment)
        return urlunparse(pieces)
    #----------------------------------------------------------------------
    def _format_tags(self, tag):
        """attempts to format tags by stripping out the {text} from the keys"""
        import re
        regex = r".*\}(.*)"
        matches = re.search(regex,tag)
        if matches:
            return matches.groups()[0]
        return tag
    #----------------------------------------------------------------------
    def _xml_to_dictionary(self, t):
        """ converts the xml to a dictionary object (recursivly)"""
        import json
        from collections import defaultdict
        d = {self._format_tags(t.tag): {} if t.attrib else None}
        children = list(t)
        if children:
            dd = defaultdict(list)
            for dc in map(self._xml_to_dictionary, children):
                for k, v in dc.items():
                    dd[self._format_tags(k)].append(v)
            d = {self._format_tags(t.tag): {self._format_tags(k): v[0] if len(v) == 1 else v for k, v in dd.items()}}
        if t.attrib:
            d[self._format_tags(t.tag)].update([('@' + self._format_tags(k), v) for k, v in t.attrib.items()])
        if t.text:
            text = t.text.strip()
            if children or t.attrib:
                if text:
                    d[self._format_tags(t.tag)]['#text'] = text
            else:
                d[self._format_tags(t.tag)] = text
        removals = ["{http://www.opengis.net/wmts/1.0}",
                    "{http://www.opengis.net/ows/1.1}",
                    "{http://www.w3.org/1999/xlink}"]
        d = json.dumps(d)
        for remove in removals:
            d = d.replace(remove, "")
        return json.loads(d)
    #----------------------------------------------------------------------
    @property
    def _lyr_json(self):
        """Represents the MapView's widget JSON format"""
        return {
            "id" : self._id,
            "title" : self._title or "WMTS Layer",
            "url" : self._url,
            "version" : self._version,
            "minScale" : self.scale[0],
            "maxScale" : self.scale[1],
            "opacity" : self.opacity,
            "type" : self._type
        }

    #----------------------------------------------------------------------
    @property
    def __text__(self):
        """creates the item's text properties"""
        
        layer = None
        tile_matrix = None

        if isinstance(self.properties.Capabilities.Contents.Layer, (list, tuple)):
            layer = self.properties.Capabilities.Contents.Layer[0]
            tile_matrix = self.properties.Capabilities.Contents.TileMatrixSet[0]
        elif isinstance(self.properties.Capabilities.Contents.Layer, (dict, PropertyMap)):
            layer = self.properties.Capabilities.Contents.Layer
            tile_matrix = self.properties.Capabilities.Contents.TileMatrixSet                
        else:
            raise ValueError("Could not parse the results properly.")
    
        url_template = (layer
                        .ResourceURL['@template']
                        .replace("{TileMatrix}", "{level}")
                        .replace("{Style}", layer.Style.Identifier)
                        .replace("{TileRow}", "{row}")
                        .replace("{TileCol}", "{col}")
                        .replace("{TileMatrixSet}",
                                 tile_matrix.Identifier)
                        )
        fullExtent = [float(coord) for coord in \
                      layer.BoundingBox.LowerCorner.strip().split(" ")] + [float(coord) for coord in \
                                                                           layer.BoundingBox.UpperCorner.strip().split(" ")]
        lods = []
        WMTS_DPI = 90.71428571428571
        for l in tile_matrix.TileMatrix:
            lods.append({
                "level": int(l.Identifier),
                "levelValue": l.Identifier,
                "resolution": float(l.ScaleDenominator) * 0.00028,
                "scale": float(l.ScaleDenominator) * WMTS_DPI / 96
            })
        return {
            "templateUrl": url_template,
            "copyright": "",
            "fullExtent": {
                "xmin": fullExtent[0],
                "ymin": fullExtent[1],
                "xmax": fullExtent[2],
                "ymax": fullExtent[3],
                "spatialReference": {
                    "wkid": int(layer.BoundingBox['@crs'].split(":")[-1])
                }
                },
            "tileInfo": {
                "rows": 256,
                "cols": 256,
                "dpi": 96,
                "origin": {
                    "x": (fullExtent[2] + fullExtent[0])/2,
                    "y": (fullExtent[3] + fullExtent[1])/2,
                    "spatialReference": {
                        "wkid": int(layer.BoundingBox['@crs'].split(":")[-1])
                    }
                },
                "spatialReference": {
                    "wkid": int(layer.BoundingBox['@crs'].split(":")[-1])
                },
                "lods": lods
            },
            "wmtsInfo": {
                "url": self._url,
                "layerIdentifier": layer.Title,
                "tileMatrixSet": [tile_matrix.Identifier]
            }
        }

    @property
    def _operational_layer_json(self):
        """Represents the WebMap's JSON format"""
        return self.__text__